# -*- coding: utf-8 -*-
"""Provide functionality to read optical constants in format provided by `refractiveindex.info <https://refractiveindex.info/>`_ website"""

from scipy.interpolate import interp1d
import numpy as np
import yaml

def read_refractive_index_from_yaml(filename, vacuum_wavelength, units="mkm", kind=1):
    """Read optical constants in format provided by refractiveindex.info website.

    Args:
            filename (str): path and file name for yaml data
                            downloaded from refractiveindex.info
            vacuum_wavelength (float or np.array): wavelengths where refractive
                                                   index data is needed
            units (str): units for wavelength. currently, microns ('mkm' or 'um')
                         and nanometers ('nm') can be selected
            kind (int): order of interpolation
    
    Returns:
            A pair (or np.array of pairs) of wavelength and
            corresponding refractive index (complex)
    """
    if units == "nm": 
        factor = 1000
    elif units in ("mkm", "um"):
        factor = 1
    else:
        raise NotImplementedError("Converting wavelength into '"+units
                                  +"' units for refractive index data"
                                  +" was not implemented.")

    with open(filename) as f:
        the_file = yaml.load(f, yaml.SafeLoader)['DATA'][0]
    data_type = the_file['type']
    if data_type != 'tabulated nk':
        raise NotImplementedError("Input data type '"+data_type
                                  +"' available in file "+filename
                                  +" was not implemented.")
    data = the_file['data'].splitlines()
    data_split = []
    for wl in data:
        data_split.append(wl.split())
    data_num = []
    for wl in data_split:
        record = []
        for val in wl:
            record.append(float(val))
        data_num.append(record)
    data_np = np.array(data_num)
    data_wl = data_np[:,0]*factor
    index_re = data_np[:,1]
    index_im = data_np[:,2]
    f_re = interp1d(data_wl, index_re, kind=kind)
    f_im = interp1d(data_wl, index_im, kind=kind)
    data_out = np.transpose(np.vstack((vacuum_wavelength, f_re(vacuum_wavelength)+f_im(vacuum_wavelength)*1j)))
    if len(data_out) == 1:
        return data_out[0]
    return data_out



"""Manage post processing steps to evaluate the scattered electric field"""

import sys
from tqdm import tqdm
import smuthi.fields as flds
import smuthi.fields.expansions as fldex
import smuthi.fields.transformations as trf
import numpy as np
from smuthi.layers import layersystem_response_matrix
from smuthi.fields.vector_wave_functions import plane_vector_wave_function
from smuthi.fields.transformations import transformation_coefficients_vwf

def scattered_field_piecewise_expansion(vacuum_wavelength, particle_list, layer_system, k_parallel='default',
                                        azimuthal_angles='default', angular_resolution=None, layer_numbers=None):
    """Compute a piecewise field expansion of the scattered field.

    Args:
        vacuum_wavelength (float):                  vacuum wavelength
        particle_list (list):                       list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):   stratified medium
        k_parallel (numpy.ndarray or str):          in-plane wavenumbers array.
                                                    if 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        azimuthal_angles (numpy.ndarray or str):    azimuthal angles array
                                                    if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):                 If provided, angular arrays are generated with this angular
                                                    resolution over the default angular range
        layer_numbers (list):                       if specified, append only plane wave expansions for these layers


    Returns:
        scattered field as smuthi.field_expansion.PiecewiseFieldExpansion object

    """
    if layer_numbers is None:
        layer_numbers = range(layer_system.number_of_layers())

    if type(k_parallel) == str and k_parallel == 'default':
        k_parallel = flds.default_Sommerfeld_k_parallel_array

    if angular_resolution is not None:
        azimuthal_angles, _ = flds.angular_arrays(angular_resolution)
    if type(azimuthal_angles) == str and azimuthal_angles == 'default':
        azimuthal_angles = flds.default_azimuthal_angles

    sfld = fldex.PiecewiseFieldExpansion()
    for i in layer_numbers:
        # layer mediated scattered field ---------------------------------------------------------------------------
        k = flds.angular_frequency(vacuum_wavelength) * layer_system.refractive_indices[i]
        ref = [0, 0, layer_system.reference_z(i)]
        vb = (layer_system.lower_zlimit(i), layer_system.upper_zlimit(i))
        pwe_up = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='upgoing',
                                          reference_point=ref, lower_z=vb[0], upper_z=vb[1])
        pwe_down = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles,
                                            kind='downgoing', reference_point=ref, lower_z=vb[0], upper_z=vb[1])
        
        for particle in tqdm(particle_list, desc='Scatt. field expansion (%i)'%i, file=sys.stdout,
                                        bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}'):
            add_up, add_down = trf.swe_to_pwe_conversion(particle.scattered_field, k_parallel, azimuthal_angles,
                                                         layer_system, i, True)
            pwe_up = pwe_up + add_up
            pwe_down = pwe_down + add_down

            pwe_up.validity_conditions.append(particle.is_outside)
            pwe_down.validity_conditions.append(particle.is_outside)

        # in bottom_layer, suppress upgoing waves, and in top layer, suppress downgoing waves
        if i > 0:
            sfld.expansion_list.append(pwe_up)
        if i < layer_system.number_of_layers()-1:
            sfld.expansion_list.append(pwe_down)

    # direct field ---------------------------------------------------------------------------------------------
    for particle in particle_list:
        sfld.expansion_list.append(particle.scattered_field)
        sfld.validity_conditions.append(particle.is_outside)

    return sfld


def scattered_field_pwe(vacuum_wavelength, particle_list, layer_system, layer_number, k_parallel='default',
                        azimuthal_angles='default', angular_resolution=None, include_direct=True, include_layer_response=True,
                        only_l=None, only_m=None, only_pol=None, only_tau=None):
    """Calculate the plane wave expansion of the scattered field of a set of particles.

    Args:
        vacuum_wavelength (float):          Vacuum wavelength (length unit)
        particle_list (list):               List of Particle objects
        layer_system (smuthi.layers.LayerSystem):  Stratified medium
        layer_number (int):                 Layer number in which the plane wave expansion should be valid
        k_parallel (numpy.ndarray or str):      in-plane wavenumbers array.
                                                if 'default', use smuthi.fields.default_Sommerfeld_k_parallel_array
        azimuthal_angles (numpy.ndarray or str):azimuthal angles array
                                                if 'default', use smuthi.fields.default_azimuthal_angles
        angular_resolution (float):             If provided, angular arrays are generated with this angular resolution
                                                over the default angular range
        include_direct (bool):                  If True, include the direct scattered field
        include_layer_response (bool):          If True, include the layer system response
        only_pol (int):  if set to 0 or 1, only this plane wave polarization (0=TE, 1=TM) is considered
        only_tau (int):  if set to 0 or 1, only this spherical vector wave polarization (0 — magnetic, 1 — electric) is
                         considered
        only_l (int):    if set to positive number, only this multipole degree is considered
        only_m (int):    if set to non-negative number, only this multipole order is considered

    Returns:
        A tuple of PlaneWaveExpansion objects for upgoing and downgoing waves.
    """

    sys.stdout.write('Evaluating scattered field plane wave expansion in layer number %i ...\n'%layer_number)
    sys.stdout.flush()

    if type(k_parallel) == str and k_parallel == 'default':
        k_parallel = flds.default_Sommerfeld_k_parallel_array

    if angular_resolution is not None:
        azimuthal_angles, _ = flds.angular_arrays(angular_resolution)
    if type(azimuthal_angles) == str and azimuthal_angles == 'default':
        azimuthal_angles = flds.default_azimuthal_angles

    omega = flds.angular_frequency(vacuum_wavelength)
    k = omega * layer_system.refractive_indices[layer_number]
    z = layer_system.reference_z(layer_number)
    vb = (layer_system.lower_zlimit(layer_number), layer_system.upper_zlimit(layer_number))
    pwe_up = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='upgoing',
                                      reference_point=[0, 0, z], lower_z=vb[0], upper_z=vb[1])
    pwe_down = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='downgoing',
                                        reference_point=[0, 0, z], lower_z=vb[0], upper_z=vb[1])

    for iS, particle in enumerate(tqdm(particle_list, desc='Scatt. field pwe          ', file=sys.stdout,
                                        bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}')):

        i_iS = layer_system.layer_number(particle.position[2])

        # direct contribution
        if i_iS == layer_number and include_direct:
            pu, pd = trf.swe_to_pwe_conversion(swe=particle.scattered_field, k_parallel=k_parallel,
                                               azimuthal_angles=azimuthal_angles, layer_system=layer_system,
                                               only_l=only_l, only_m=only_m, only_pol=only_pol, only_tau=only_tau)
            pwe_up = pwe_up + pu
            pwe_down = pwe_down + pd

        # layer mediated contribution
        if include_layer_response:
            pu, pd = trf.swe_to_pwe_conversion(swe=particle.scattered_field, k_parallel=k_parallel,
                                               azimuthal_angles=azimuthal_angles, layer_system=layer_system,
                                               layer_number=layer_number, layer_system_mediated=True,
                                               only_l=only_l, only_m=only_m, only_pol=only_pol, only_tau=only_tau)
            pwe_up = pwe_up + pu
            pwe_down = pwe_down + pd

    return pwe_up, pwe_down


def evaluate_scattered_field_stat_phase_approx(x, y, z, vacuum_wavelength, particle_list, layer_system):
    """Evaluate the scattered electric field for N particles on a substrate. The substrate reflection is evaluated
    by means of the stationary phase approximation, as presented in
    "A quick way to approximate a Sommerfeld-Weyl_type Sommerfeld integral" by W.C. Chew (1988).

    See also the technical note "Usage of the stationary phase approximation in SMUTHI" by A. Egel (2020)

    The stationary phase approximation is expected to yield good results for field points far away from the particles.

    ********************************************************************************************************************
    Note: This function assumes that the particles are located in the upper layer of a two-layer system (particles on
    substrate). For other cases, this function does not apply.
    ********************************************************************************************************************

    Args:
        x (float or numpy.ndarray):                                  x-coordinates of query points
        y (float or numpy.ndarray):                                  y-coordinates of query points
        z (float or numpy.ndarray):                                  z-coordinates of query points
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        particle_list (list):                               List of Particle objects
        layer_system (smuthi.layers.LayerSystem):           Stratified medium

    Returns:
        Tuple of (E_x, E_y, E_z) numpy.ndarray objects with the Cartesian coordinates of complex electric field.
    """
    x = np.array(x, ndmin=1)
    y = np.array(y, ndmin=1)
    z = np.array(z, ndmin=1)

    # assert a two layer system
    assert (len(layer_system.refractive_indices) == 2)

    # assert that all particles are in top layer (this function applies only to this special case)
    for particle in particle_list:
        assert (layer_system.layer_number(particle.position[2]) == 1)

    # assert that all query points are in top layer (this function applies only to this special case)
    for zi in z:
        assert (layer_system.layer_number(zi) == 1)

    ex = np.zeros(x.shape, dtype=complex)
    ey = np.zeros(x.shape, dtype=complex)
    ez = np.zeros(x.shape, dtype=complex)

    k = layer_system.wavenumber(1, vacuum_wavelength)

    for particle in tqdm(particle_list, desc='Scatt. field st. ph. appr.', file=sys.stdout,
                                        bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}'):

        # direct scattered field
        exdir, eydir, ezdir = particle.scattered_field.electric_field(x, y, z)
        ex += exdir
        ey += eydir
        ez += ezdir

        # layer-reflected scattered field
        xtilde = x - particle.position[0]
        ytilde = y - particle.position[1]
        ztilde = z + particle.position[2]
        rtilde = np.sqrt(xtilde**2 + ytilde**2 + ztilde**2)

        # stationary phase wave-vector
        k0x = k * xtilde / rtilde
        k0y = k * ytilde / rtilde
        k0z = k * ztilde / rtilde
        k0par = np.hypot(k0x, k0y)
        alpha0 = np.arctan2(k0y, k0x)

        prefac = -1j * np.exp(1j * k * rtilde) / (k * rtilde)

        for pol in range(2):
            L = layersystem_response_matrix(pol=pol,
                                            layer_d=layer_system.thicknesses,
                                            layer_n=layer_system.refractive_indices,
                                            kpar=k0par,
                                            omega=flds.angular_frequency(vacuum_wavelength),
                                            fromlayer=1,
                                            tolayer=1)
            fresnel_coeff = L[0, 1]
            pvwf_direction = plane_vector_wave_function(0, 0, 0, k0par, alpha0, k0z, pol)
            print(pvwf_direction[2][-1])

            for tau in range(2):
                for m in range(-particle.scattered_field.m_max, particle.scattered_field.m_max + 1):
                    eima = np.exp(1j * m * alpha0)
                    for l in range(max(1, abs(m)), particle.scattered_field.l_max + 1):
                        b = particle.scattered_field.coefficients_tlm(tau, l, m)
                        Bnj = transformation_coefficients_vwf(tau=tau, l=l, m=m, pol=pol, kp=k0par, kz=-k0z)
                        scalar_fac = prefac * fresnel_coeff * b * eima * Bnj
                        ex += scalar_fac * pvwf_direction[0]
                        ey += scalar_fac * pvwf_direction[1]
                        ez += scalar_fac * pvwf_direction[2]

    return ex, ey, ez



# -*- coding: utf-8 -*-
"""This module contains the vector wave functions and their transformations."""
 
import numpy as np
import smuthi.utility.math as sf


def plane_vector_wave_function(x, y, z, kp, alpha, kz, pol):
    r"""Electric field components of plane wave (PVWF), see section 2.3.1 of
    [Egel 2018 diss].

    .. math::
        \mathbf{\Phi}_j = \exp ( \mathrm{i} \mathbf{k} \cdot \mathbf{r} ) \hat{ \mathbf{e} }_j

    with :math:`\hat{\mathbf{e}}_0` denoting the unit vector in azimuthal direction ('TE' or 's' polarization),
    and :math:`\hat{\mathbf{e}}_1` denoting the unit vector in polar direction ('TM' or 'p' polarization).

    The input arrays should have one of the following dimensions:

        - x,y,z: (N x 1) matrix
        - kp,alpha,kz: (1 x M) matrix
        - Ex, Ey, Ez: (M x N) matrix

    or

        - x,y,z: (M x N) matrix
        - kp,alpha,kz: scalar
        - Ex, Ey, Ez: (M x N) matrix

    Args:
        x (numpy.ndarray): x-coordinate of position where to test the field (length unit)
        y (numpy.ndarray): y-coordinate of position where to test the field
        z (numpy.ndarray): z-coordinate of position where to test the field
        kp (numpy.ndarray): parallel component of k-vector (inverse length unit)
        alpha (numpy.ndarray): azimthal angle of k-vector (rad)
        kz (numpy.ndarray): z-component of k-vector (inverse length unit)
        pol (int): Polarization (0=TE, 1=TM)

    Returns:
        - x-coordinate of PVWF electric field (numpy.ndarray)
        - y-coordinate of PVWF electric field (numpy.ndarray)
        - z-coordinate of PVWF electric field (numpy.ndarray)
    """
    k = np.sqrt(kp**2 + kz**2)
    kx = kp * np.cos(alpha)
    ky = kp * np.sin(alpha)

    scalar_wave = np.exp(1j * (kx * x + ky * y + kz * z))

    if pol == 0:
        Ex = -np.sin(alpha) * scalar_wave
        Ey = np.cos(alpha) * scalar_wave
        Ez = np.zeros_like(scalar_wave)
    elif pol == 1:
        Ex = np.cos(alpha) * kz / k * scalar_wave
        Ey = np.sin(alpha) * kz / k * scalar_wave
        Ez = -kp / k * scalar_wave
    else:
        raise ValueError('Polarization must be 0 (TE) or 1 (TM)')

    return Ex, Ey, Ez


def spherical_vector_wave_function(x, y, z, k, nu, tau, l, m):
    """Electric field components of spherical vector wave function (SVWF). The conventions are chosen according to
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_
    See also section 2.3.2 of [Egel 2018 diss].

    Args:
        x (numpy.ndarray):      x-coordinate of position where to test the field (length unit)
        y (numpy.ndarray):      y-coordinate of position where to test the field
        z (numpy.ndarray):      z-coordinate of position where to test the field
        k (float or complex):   wavenumber (inverse length unit)
        nu (int):               1 for regular waves, 3 for outgoing waves
        tau (int):              spherical polarization, 0 for spherical TE and 1 for spherical TM
        l (int):                l=1,... multipole degree (polar quantum number)
        m (int):                m=-l,...,l multipole order (azimuthal quantum number)

    Returns:
        - x-coordinate of SVWF electric field (numpy.ndarray)
        - y-coordinate of SVWF electric field (numpy.ndarray)
        - z-coordinate of SVWF electric field (numpy.ndarray)
    """
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    r = np.sqrt(x**2 + y**2 + z**2)
    non0 = np.logical_not(r == 0)
    
    theta = np.zeros(x.shape)
    theta[non0] = np.arccos(z[non0] / r[non0])
    phi = np.arctan2(y, x)

    # unit vector in r-direction

    er_x = np.zeros(x.shape)
    er_y = np.zeros(x.shape)
    er_z = np.ones(x.shape)

    er_x[non0] = x[non0] / r[non0]
    er_y[non0] = y[non0] / r[non0]
    er_z[non0] = z[non0] / r[non0]

    # unit vector in theta-direction
    eth_x = np.cos(theta) * np.cos(phi)
    eth_y = np.cos(theta) * np.sin(phi)
    eth_z = -np.sin(theta)

    # unit vector in phi-direction
    eph_x = -np.sin(phi)
    eph_y = np.cos(phi)
    eph_z = np.zeros_like(x)

    cos_thet = np.cos(theta)
    sin_thet = np.sin(theta)
    plm_list, pilm_list, taulm_list = sf.legendre_normalized(cos_thet, sin_thet, l)
    plm = plm_list[l, abs(m)]
    pilm = pilm_list[l, abs(m)]
    taulm = taulm_list[l, abs(m)]

    kr = k * r
    if nu == 1:
        bes = sf.spherical_bessel(l, kr)
        dxxz = sf.dx_xj(l, kr)
        bes_kr = np.zeros(kr.shape, dtype=complex)
        dxxz_kr = np.zeros(kr.shape, dtype=complex)
        zero = (r==0)
        nonzero = np.logical_not(zero)
        bes_kr[zero] = (l == 1) / 3
        dxxz_kr[zero] = (l == 1) * 2 / 3
        bes_kr[nonzero] = (bes[nonzero] / kr[nonzero])
        dxxz_kr[nonzero] = (dxxz[nonzero] / kr[nonzero])
    elif nu == 3:
        bes = sf.spherical_hankel(l, kr)
        dxxz = sf.dx_xh(l, kr)
        bes_kr = bes / kr
        dxxz_kr = dxxz / kr
    else:
        raise ValueError('nu must be 1 (regular SVWF) or 3 (outgoing SVWF)')

    eimphi = np.exp(1j * m * phi)
    prefac = 1 / np.sqrt(2 * l * (l + 1))
    if tau == 0:
        Ex = prefac * bes * (1j * m * pilm * eth_x - taulm * eph_x) * eimphi
        Ey = prefac * bes * (1j * m * pilm * eth_y - taulm * eph_y) * eimphi
        Ez = prefac * bes * (1j * m * pilm * eth_z - taulm * eph_z) * eimphi
    elif tau == 1:
        Ex = prefac * (l * (l + 1) * bes_kr * plm * er_x + dxxz_kr * (taulm * eth_x + 1j * m * pilm * eph_x)) * eimphi
        Ey = prefac * (l * (l + 1) * bes_kr * plm * er_y + dxxz_kr * (taulm * eth_y + 1j * m * pilm * eph_y)) * eimphi
        Ez = prefac * (l * (l + 1) * bes_kr * plm * er_z + dxxz_kr * (taulm * eth_z + 1j * m * pilm * eph_z)) * eimphi
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return Ex, Ey, Ez



"""Manage post processing steps to evaluate the electric field inside a sphere"""

import smuthi.fields as flds
import smuthi.fields.expansions as fldex
import smuthi.linearsystem.tmatrix.t_matrix as tmt


def internal_field_piecewise_expansion(vacuum_wavelength, particle_list, layer_system):
    """Compute a piecewise field expansion of the internal field of spheres.

    Args:
        vacuum_wavelength (float):                  vacuum wavelength
        particle_list (list):                       list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):   stratified medium

    Returns:
        internal field as smuthi.field_expansion.PiecewiseFieldExpansion object

    """
    intfld = fldex.PiecewiseFieldExpansion()

    for particle in particle_list:
        if type(particle).__name__ == 'Sphere':
            i_part = layer_system.layer_number(particle.position[2])
            k_medium = flds.angular_frequency(vacuum_wavelength) * layer_system.refractive_indices[i_part]
            k_particle = flds.angular_frequency(vacuum_wavelength) * particle.refractive_index

            internal_field = fldex.SphericalWaveExpansion(
                k_particle,
                l_max=particle.l_max,
                m_max=particle.m_max,
                kind='regular',
                reference_point=particle.position)

            internal_field.validity_conditions.append(particle.is_inside)

            for m in range(-particle.m_max, particle.m_max+1):
                for l in range(max(1, abs(m)), particle.l_max+1):
                    for tau in range(2):
                        n = flds.multi_to_single_index(tau, l, m, particle.l_max, particle.m_max)
                        b_to_c = (tmt.internal_mie_coefficient(tau, l, k_medium, k_particle, particle.radius)
                                  / tmt.mie_coefficient(tau, l, k_medium, k_particle, particle.radius))
                        internal_field.coefficients[n] = particle.scattered_field.coefficients[n] * b_to_c

            intfld.expansion_list.append(internal_field)
            particle.internal_field = internal_field

    return intfld



"""This module contains helper functions to evaluate the coupling between a 
single particle and a periodic particle arrangement."""

import smuthi.fields as flds
from numba import njit
import numpy as np


def a5b5_lookup(lmax1, lmax2, mmax1, mmax2):
    """ Prepare lookup table of smuthi.vector_wave_functions.ab5_coefficients().
        Allows Numba support for the evaluation of the coupling matrix of periodic particle arrangements.
    Args:
        lmax1 (int):        l=1,...lmax: Original wave's SVWF multipole degree
        mmax1 (int):        m=-mmax,...,mmax: Original wave's SVWF multipole order
        lmax2 (int):        l=1,...lmax: Partial wave's SVWF multipole degree
        mmax2 (int):        m=-mmax,...,mmax: Partial wave's SVWF multipole order
    Returns: 
        Lookup table for a5 and b5 coefficients. 
        The respective coefficients can be called via [a=0 b=1, l1 - 1, l2 - 1, m1 + mmax2, m2 + mmax2, p].
    """

    a5b5mat = np.zeros([2, lmax1, lmax2, 2*mmax1+1, 2*mmax2+1, lmax1+lmax2+1], np.complex128)
    for l1 in range(1, lmax1 + 1):
        for l2 in range(1, lmax2 + 1):
            for m1 in range(-min(mmax1, l1), min(mmax1, l1) + 1):
                for m2 in range(-min(mmax2, l2), min(mmax2, l2) + 1):
                    M = m2 - m1
                    for L in range(max(abs(l1 - l2), abs(M)), l1 + l2 + 1): # if L < abs(M) then p=0   
                        a5b5mat[:, l1 - 1, l2 - 1, m1 + mmax1, m2 + mmax2, L] = flds.transformations.ab5_coefficients(l2, m2, l1, m1, L)
    return a5b5mat


@njit() 
def transformation_coefficients_vwf(tau, l, m, pol, kp=None, kz=None, pilm=None, taulm=None, dagger=False):
    r""" Numba compatible copy of smuthi.vector_wave_functions.transformation_coefficients_vwf()
    Transformation coefficients B to expand SVWF in PVWF and vice versa:

    .. math::
        B_{\tau l m,j}(x) = -\frac{1}{\mathrm{i}^{l+1}} \frac{1}{\sqrt{2l(l+1)}} (\mathrm{i} \delta_{j1} + \delta_{j2})
        (\delta_{\tau j} \tau_l^{|m|}(x) + (1-\delta_{\tau j} m \pi_l^{|m|}(x))

    For the definition of the :math:`\tau_l^m` and :math:`\pi_l^m` functions, see
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_

    Args:
        tau (int):          SVWF polarization, 0 for spherical TE, 1 for spherical TM
        l (int):            l=1,... SVWF multipole degree
        m (int):            m=-l,...,l SVWF multipole order
        pol (int):          PVWF polarization, 0 for TE, 1 for TM
        kp (numpy array):         PVWF in-plane wavenumbers
        kz (numpy array):         complex numpy-array: PVWF out-of-plane wavenumbers
        pilm (numpy array):       pilm[l, m, *ct.shape]
        taulm (numpy array):      tauilm[l, m, *ct.shape]
        dagger (bool):      switch on when expanding PVWF in SVWF and off when expanding SVWF in PVWF

    Returns:
        Transformation coefficient as array (size like kp).
    """
    # numba cannot handle these "is None" questions
    # if pilm is None: 
    #     k = np.sqrt(kp**2 + kz**2)
    #     ct = kz / k
    #     st = kp / k
    #     _, pilm, taulm = sf.legendre_normalized_numbed(ct, st, l)

    if tau == pol:
        sphfun = taulm[l, abs(m)]
    else:
        sphfun = m * pilm[l, abs(m)]

    if dagger:
        if pol == 0:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (-1j)
        elif pol == 1:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')
    else:
        if pol == 0:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (1j)
        elif pol ==1:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')

    B = prefac * sphfun
    return B


@njit()
def index_block(i, lmax_array, mmax_array):
    """ Numba compatible copy of smuthi.linearsystem.linear_system.System_matrix.index_block()
    Args:
        i (int):                        number of particle
        lmax_array (numpy.ndarray):     lmax of each particle 
        mmax_array (numpy.ndarray):     mmax of each particle
    Returns:
        Indices that correspond to the coefficients for that particle   
    """
    blocksizes = np.zeros((i + 1), np.int32)
    for idx in range(i + 1):
        blocksizes[idx] = flds.blocksize(lmax_array[idx], mmax_array[idx])
    return (np.sum(blocksizes[:i]), np.sum(blocksizes))


###############################################################################
#                  Numba compatible smuthi.layer functions                    #                               
###############################################################################
@njit()
def fresnel_r(pol, kz1, kz2, n1, n2):
    """ Numba compatible copy of smuthi.layers.fresnel_r()
        Fresnel reflection coefficient.
    Args:
        pol (int):              polarization (0=TE, 1=TM)
        kz1 (float or array):   incoming wave's z-wavenumber (k*cos(alpha1))
        kz2 (float or array):   transmitted wave's z-wavenumber (k*cos(alpha2))
        n1 (float or complex):  first medium's complex refractive index (n+ik)
        n2 (float or complex):  second medium's complex refractive index (n+ik)
    Returns:
        Complex Fresnel reflection coefficient (float or array)
    """
    if pol == 0:
        return (kz1 - kz2) / (kz1 + kz2)
    else:
        return (n2 ** 2 * kz1 - n1 ** 2 * kz2) / (n2 ** 2 * kz1 + n1 ** 2 * kz2)
    

@njit()    
def fresnel_t(pol, kz1, kz2, n1, n2):
    """ Numba compatible copy of smuthi.layers.fresnel_t()
        Fresnel transmission coefficient.
    Args:
        pol (int):              polarization (0=TE, 1=TM)
        kz1 (float or array):   incoming wave's z-wavenumber (k*cos(alpha1))
        kz2 (float or array):   transmitted wave's z-wavenumber (k*cos(alpha2))
        n1 (float or complex):  first medium's complex refractive index (n+ik)
        n2 (float or complex):  second medium's complex refractive index (n+ik)
    Returns:
        Complex Fresnel transmission coefficient (float or array)
    """
    if pol == 0:
        return 2 * kz1 / (kz1 + kz2)
    else:
        return 2 * n1 * n2 * kz1 / (n2 ** 2 * kz1 + n1 ** 2 * kz2)    


@njit()
def interface_transition_matrix(pol, kz1, kz2, n1, n2):
    """ Numba compatible copy of smuthi.layers.interface_transition_matrix()
        Interface transition matrix to be used in the Transfer matrix algorithm.
    Args:
        pol (int):              polarization (0=TE, 1=TM)
        kz1 (float or complex):   incoming wave's z-wavenumber (k*cos(alpha1))
        kz2 (float or complex):   transmitted wave's z-wavenumber (k*cos(alpha2))
        n1 (float or complex):  first medium's complex refractive index (n+ik)
        n2 (float or complex):  second medium's complex refractive index (n+ik)
    Returns:
        Interface transition matrix as 2x2 numpy array 
    """
    t = fresnel_t(pol, kz1, kz2, n1, n2)
    r = fresnel_r(pol, kz1, kz2, n1, n2)
    return 1 / t * np.array(((1, r), (r, 1)), np.complex128)


@njit()
def layer_propagation_matrix(kz, d):
    """ Numba compatible copy of smuthi.layers.layer_propagation_matrix()
        Layer propagation matrix to be used in the Transfer matrix algorithm.
    Args:
        kz (float or complex):  z-wavenumber (k*cos(alpha))
        d  (float):             thickness of layer
    Returns:
        Layer propagation matrix as 2x2 numpy array 
    """
    arr = np.array(((np.exp(-1j * kz * d), 0), (0, np.exp(1j * kz * d))), np.complex128)    
    # remove inf (and nan), inf appears for thick layers and evanescent waves
    if np.isinf(arr[0, 0]):
        arr[0, 0] = 1.79769313e+307
        arr[1, 1] = 1e-308       
    return arr
    

@njit()
def layersystem_transfer_matrix(pol, layer_d, layer_n, kpar, omega):
    """ Numba compatible copy of smuthi.layers.layersystem_transfer_matrix()
        Transfer matrix of a planarly layered medium.
    Args:
        pol (int):                  polarization(0=TE, 1=TM)
        layer_d (numpy.ndarray):    layer thicknesses (float64)
        layer_n (numpy.ndarray):    complex layer refractive indices
        kpar (complex):               in-plane wavenumber
        omega (float):              angular frequency in units of c=1: omega=2*pi/lambda
    Returns:
        Transfer matrix as 2x2 numpy array 
    """
    layer_kz = []
    for n in layer_n:
        kz = np.sqrt((omega * n) ** 2 - kpar ** 2 + 0j)
        if kz.imag < 0:
            kz = -kz
        layer_kz.append(kz)
    tmat = np.array(((1, 0), (0, 1)), np.complex128)
    for i in range(layer_d.shape[0] - 1):
        dmat = interface_transition_matrix(pol, layer_kz[i], layer_kz[i + 1], layer_n[i], layer_n[i + 1])
        pmat = layer_propagation_matrix(layer_kz[i], layer_d[i])
        tmat = np.dot(tmat, np.dot(pmat, dmat))
    return tmat


@njit()
def layersystem_scattering_matrix(pol, layer_d, layer_n, kpar, omega):
    """ Numba compatible copy of smuthi.layers.layersystem_scattering_matrix()
        Scattering matrix of a planarly layered medium.
    Args:
        pol (int):                  polarization(0=TE, 1=TM)
        layer_d (numpy.ndarray):    layer thicknesses (float64)
        layer_n (numpy.ndarray):    complex layer refractive indices
        kpar (complex):             in-plane wavenumber
        omega (float):              angular frequency in units of c=1: omega=2*pi/lambda
    Returns:
        Scattering matrix as 2x2 numpy array 
    """
    layer_kz = []
    for n in layer_n:
        kz = np.sqrt((omega * n) ** 2 - kpar ** 2 + 0j)
        if kz.imag < 0:
            kz = -kz
        layer_kz.append(kz)
    smat = np.array(((1, 0), (0, 1)), np.complex128)
    for i in range(layer_d.shape[0] - 1):
        dmat = interface_transition_matrix(pol, layer_kz[i], layer_kz[i + 1], layer_n[i], layer_n[i + 1])
        pmat = layer_propagation_matrix(layer_kz[i], layer_d[i])
        tmat = np.dot(pmat, dmat)
        s11 = smat[0, 0] / (tmat[0, 0] - smat[0, 1] * tmat[1, 0])
        s12 = (smat[0, 1] * tmat[1, 1] - tmat[0, 1]) / (tmat[0, 0] - smat[0, 1] * tmat[1, 0])
        s21 = smat[1, 1] * tmat[1, 0] * s11 + smat[1, 0]
        s22 = smat[1, 1] * tmat[1, 0] * s12 + smat[1, 1] * tmat[1, 1]
        smat = np.array(((s11, s12), (s21, s22)), np.complex128)
    return smat


@njit()
def layersystem_response_matrix(pol, layer_d, layer_n, kpar, omega, fromlayer, tolayer):
    """ Numba compatible copy of smuthi.layers.layersystem_response_matrix()
        Layer system response matrix of a planarly layered medium.
    Args:
        pol (int):                          polarization(0=TE, 1=TM)
        layer_d (numpy.ndarray):            layer thicknesses (float64)
        layer_n (numpy.ndarray):            complex layer refractive indices
        kpar (numpy.ndarray):               in-plane wavenumber
        omega (float):                      angular frequency in units of c=1: omega=2*pi/lambda
        fromlayer (int):                    number of layer where the excitation is located
        tolayer (int):                      number of layer where the response is evaluated
    Returns:
        Layer system response matrix as 2x2xN array with kpar is array of len = N.
    """
    
    # if kpar is an array of only one value, the function returns a (2x2x1) matrix /different to the original function
    # necessary since numba expects the function to always return arrays of the same number of dimensions
    
    shape = len(kpar)
    if not shape == 1: # is kpar an array? then use recursive call to fill an 2 x 2 x N ndarray
        result = np.zeros((2, 2, shape), np.complex128)
        for i in range(shape):
            kp = np.array([kpar[i]], np.complex128)
            result[:, :, i] = layersystem_response_matrix(pol, layer_d, layer_n, kp,
                                                          omega, fromlayer, tolayer).reshape(2, 2)
        return result

    layer_d_above = np.append(np.array([0], np.float64),  layer_d[fromlayer:])
    layer_n_above = np.append(np.array([layer_n[fromlayer]], np.complex128),  layer_n[fromlayer:])
    smat_above = layersystem_scattering_matrix(pol, layer_d_above, layer_n_above, kpar[0], omega)
    layer_d_below = np.append(layer_d[:fromlayer], np.array([0], np.float64))
    layer_n_below = layer_n[:fromlayer + 1]
    smat_below = layersystem_scattering_matrix(pol, layer_d_below, layer_n_below, kpar[0], omega)
    lmat = np.dot(np.linalg.inv(np.array(((1, -smat_below[0, 1]), (-smat_above[1, 0], 1)), np.complex128)),
                          np.array(((0, smat_below[0, 1]), (smat_above[1, 0], 0)), np.complex128))
    if tolayer > fromlayer:
        tmat_fromto = layersystem_transfer_matrix(pol, layer_d[fromlayer:tolayer + 1], layer_n[fromlayer:tolayer + 1],
                                                  kpar[0], omega)
        lmat = np.dot(np.linalg.inv(tmat_fromto), lmat + np.array(((1, 0), (0, 0)), np.complex128))
    elif tolayer < fromlayer:
        tmat_fromto = layersystem_transfer_matrix(pol, layer_d[tolayer:fromlayer + 1], layer_n[tolayer:fromlayer + 1],
                                                  kpar[0], omega)
        lmat = np.dot(tmat_fromto, lmat + np.array(((0, 0), (0, 1)), np.complex128))
    return lmat.reshape(2, 2, 1)


"""Manage post processing steps to evaluate power flux"""

import smuthi.fields
import numpy as np

def power_flux_through_zplane(vacuum_wavelength, z, upgoing_pwe=None, downgoing_pwe=None):
    """Evaluate time averaged power flux though a plane of z=const.

    Args:
        vacuum_wavelength (float):  Vacuum wavelength in length units.
        z (float):                  plane height z
        upgoing_pwe (PlaneWaveExpansion):   of kind "upgoing"
        downgoing_pwe (PlaneWaveExpansion): of kind "downgoing"

    Returns:
        Time averaged energy flux.

    """
    
    if not upgoing_pwe and not downgoing_pwe:
        return 0     
    if upgoing_pwe:
        if not upgoing_pwe.valid(np.array([0]), np.array([0]), np.array([z])):
            raise ValueError('Upgoing PlaneWaveExpansion is not valid.')
    if downgoing_pwe:
        if not downgoing_pwe.valid(np.array([0]), np.array([0]), np.array([z])):
            raise ValueError('Downgoing PlaneWaveExpansion is not valid.')
    
    if upgoing_pwe and downgoing_pwe:
        if upgoing_pwe.kind == downgoing_pwe.kind:
            raise ValueError('Both PlaneWaveExpansions are of same kind.')
        if not np.isclose(upgoing_pwe.k, downgoing_pwe.k):
            raise ValueError('Wavenumbers are not compatible.')
        if not all(np.isclose(upgoing_pwe.k_parallel, downgoing_pwe.k_parallel)):
            raise ValueError('In-plane wavenumbers are not compatible.')
        if not all(np.isclose(upgoing_pwe.azimuthal_angles, downgoing_pwe.azimuthal_angles)):
            raise ValueError('Azimuthal angles are not compatible.')
            
    if upgoing_pwe:
        k = upgoing_pwe.k
        kpar = upgoing_pwe.k_parallel
        alpha = upgoing_pwe.azimuthal_angles
    else:
        k = downgoing_pwe.k
        kpar = downgoing_pwe.k_parallel
        alpha = downgoing_pwe.azimuthal_angles        
    omega = smuthi.fields.angular_frequency(vacuum_wavelength)
    kz = smuthi.fields.k_z(k_parallel=kpar, k=k)
           
    if z == float('-inf') or z == float('inf'): # no evanescent waves necessary
        if upgoing_pwe:
            kpar = upgoing_pwe.k_parallel[np.where(upgoing_pwe.k_parallel <= upgoing_pwe.k)[0]]
            gpl = upgoing_pwe.coefficients[:, np.where(upgoing_pwe.k_parallel <= upgoing_pwe.k)[0], :]
        if downgoing_pwe:
            kpar = downgoing_pwe.k_parallel[np.where(downgoing_pwe.k_parallel <= downgoing_pwe.k)[0]]
            gmn = downgoing_pwe.coefficients[:, np.where(downgoing_pwe.k_parallel <= downgoing_pwe.k)[0], :]
        kz = smuthi.fields.k_z(k_parallel=kpar, k=k)
    else: # translate to z
        if upgoing_pwe:
            ejkzdz = np.exp(1j * upgoing_pwe.k_z_grid() * (z - upgoing_pwe.reference_point[2]))
            gpl = upgoing_pwe.coefficients * ejkzdz
        if downgoing_pwe:
            ejkzdz = np.exp(1j * downgoing_pwe.k_z_grid() * (z - downgoing_pwe.reference_point[2]))
            gmn = downgoing_pwe.coefficients * ejkzdz
    
    if any(kpar.imag):
        raise ValueError('complex in-plane wavenumbers are not allowed')
    
    if not upgoing_pwe:
        gpl = np.zeros(gmn.shape, dtype=complex)
    if not downgoing_pwe:
        gmn = np.zeros(gpl.shape, dtype=complex)         
    
    prefac = 2 * np.pi ** 2 / omega            
    gpl2_mn_gmn2 = np.abs(gpl) ** 2 - np.abs(gmn) ** 2
    im_gpl_t_gmnconj = (gpl * np.conjugate(gmn)).imag
    integrand = kz.real[None, : , None] * gpl2_mn_gmn2 - 2 * kz.imag[None, :, None] * im_gpl_t_gmnconj           
    return prefac * np.sum(np.trapz(np.trapz(integrand, alpha, axis=2) * kpar, kpar, axis=1))




"""
This module contains functions to compute the direct (i.e., not layer 
mediated) particle coupling coefficients.
"""

import numpy as np
import smuthi.fields as flds
import smuthi.fields.transformations as trf
from smuthi.fields.spherical_harmonic_translations import svwf_translate
import smuthi.utility.math as sma
import smuthi.utility.memoizing as memo
import scipy.optimize
import scipy.special
import sys
import warnings


def direct_coupling_block(vacuum_wavelength, receiving_particle, emitting_particle, layer_system):
    r"""Direct particle coupling matrix :math:`W` for two particles that do not have intersecting circumscribing spheres.
       This routine is explicit.
       
       To reduce computation time, this routine relies on two internal accelerations. 
       First, in most cases the number of unique maximum multipole indicies,
       :math:`(\tau, l_{max}, m_{max})`, is much less than the number of unique particles. 
       Therefore, all calculations that depend only on multipole indicies are stored in an 
       intermediate hash table. Second, Cython acceleration is used by default to leverage 
       fast looping. If the Cython files are not supported, this routine will 
       fall back on equivalent Python looping.
       
       Cython acceleration can be between 10-1,000x faster compared to the Python 
       equivalent. Speed variability depends on the number of unique multipoles indicies,
       the size of the largest multipole order, and if particles share the same z coordinate.
       

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place

    Returns:
        Direct coupling matrix block as numpy array.
    """
    
    # Note: Direct coupling is transpose of translation. 
    # The translation convention is emit = 1 -> recieve = 2
    # The transpose of this is emit = 2 -> recieve = 1
    lmax1 = int(receiving_particle.l_max) 
    mmax1 = int(receiving_particle.m_max)
    lmax2 = int(emitting_particle.l_max)
    mmax2 = int(emitting_particle.m_max)    

    
    # Check if particles are in the same layer.
    rS1 = receiving_particle.position
    rS2 = emitting_particle.position
    iS1 = layer_system.layer_number(rS1[2])
    iS2 = layer_system.layer_number(rS2[2])
    
    if (iS1 == iS2 or layer_system.is_degenerate()) and not emitting_particle == receiving_particle:
        # Initialize variables from abstract classes to be passed as simple 
        # data types
        omega = flds.angular_frequency(vacuum_wavelength)
        k = complex(omega * layer_system.refractive_indices[iS1])
        d = [rS1[i]-rS2[i] for i in range(3)]
        return svwf_translate(k, d, lmax1, mmax1, lmax2, mmax2, kind = 'outgoing to regular', threaded = False)

    else:
        blocksize1 = int(flds.blocksize(lmax1, mmax1)) # Rows of translation matrix
        blocksize2 = int(flds.blocksize(lmax2, mmax2)) # Cols of translation matrix
        # Note: Direct coupling returns zeros such that you recieve 0 on multiplication (no coupling to self..)
        # Note: Smuthi assumes that if you try to calculate direct coupling of two particles across an interface
        # that the function returns zero instead of an exception! This allows for lazy functions that do not 
        # check for the validity of direct coupling before calling. 
        return np.zeros((blocksize1, blocksize2), dtype = np.complex128, order = 'C')
    
    # To Do: lazy evaluation of direct coupling could be replaced with an explicit check?
    #raise Exception('''
        #                You are attempting to couple particles directly that exist in different layers. 
        #                Direct coupling (2D or 3D) is only sensible for particles in the same layer.
        #                You need to first deal with the effect of the layer interface!
        #                ''')


        dx = float(rS1[0] - rS2[0])
        dy = float(rS1[1] - rS2[1])
        dz = float(rS1[2] - rS2[2])

        #Note: It is allways better to use the hash tables so just do it. 
        if dz == 0:
            a5leg_array, b5leg_array = ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2)
            w = direct_coupling_block_2D_from_hash_table(blocksize1, blocksize2,
                                                        w, sph, k, dx, dy,dz,
                                                        lmax1, lmax2, mmax1, mmax2,
                                                        a5leg_array, b5leg_array,                              
                                                        threaded = False)
        else:
            a5_array, b5_array = ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2)
            w = direct_coupling_block_3D_from_hash_table(blocksize1, blocksize2,
                                                        w, sph, k, dx, dy,dz,
                                                        lmax1, lmax2, mmax1, mmax2,
                                                        a5_array, b5_array,                              
                                                        threaded = False)           
        
    return w



def direct_coupling_matrix(vacuum_wavelength, particle_list, layer_system):
    """Return the direct particle coupling matrix W for a particle collection in a layered medium.

    Args:
        vacuum_wavelength (float):                                  Wavelength in length unit
        particle_list (list of smuthi.particles.Particle obejcts:   Scattering particles
        layer_system (smuthi.layers.LayerSystem):                   The stratified medium
   
    Returns:
        Ensemble coupling matrix as numpy array.
    """
    # indices
    blocksizes = [flds.blocksize(particle.l_max, particle.m_max)
                  for particle in particle_list]

    # initialize result
    w = np.zeros((sum(blocksizes), sum(blocksizes)), dtype=complex)

    for s1, particle1 in enumerate(particle_list):
        idx1 = np.array(range(sum(blocksizes[:s1]), sum(blocksizes[:s1+1])))
        for s2, particle2 in enumerate(particle_list):
            idx2 = range(sum(blocksizes[:s2]), sum(blocksizes[:s2+1]))
            w[idx1[:, None], idx2] = direct_coupling_block(vacuum_wavelength, particle1, particle2, layer_system)

    return w


try: 
    from smuthi.utility.cython.cython_speedups import _ab5_coefficient_and_legendre_hash_table
    @memo.Memoize
    def ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2):
        return _ab5_coefficient_and_legendre_hash_table(int(lmax1), int(lmax2), int(mmax1), int(mmax2))
except:
    sys.stdout.write("Cython acceleration could not be loaded. \nFalling back on Python equivalents...")
    sys.stdout.flush()
    @memo.Memoize
    def ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2):
        r"""Creates a hash table of the elements
        :math:`a5(l_1,m_1,l_2,m_2,l_d)*P_{l_d}^{m_1-m_2}(0)` and :math:`a5(l_1,m_1,l_2,m_2,l_d)*P_{l_d}^{m_1-m_2}(0)`
        found in appendix B of [Egel 2018 diss],where a5 and b5 are the coefficients used in the evaluation of the SVWF translation
        operator and :math:`P_l^m(\cos\theta)` are the normalized associated Legendre functions. This hash table is usefull in 
        reducing computation time when calculating the coupling between two particles that exist in the same layer 
        and have the same z coordinate.


        Args:
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle

    
        Returns:
            a5leg_array (ndarray):          hash table of the elements in :math:`A` not dependent on :math:`/phi` or :math:`kd`
            b5leg_array (ndarray):          hash table of the elements in :math:`B` not dependent on :math:`/phi` or :math:`kd`          

        """
        
        a5leg_array = []
        b5leg_array = []
        #ab5leg_key = []
        
        legendre_2D = sma.legendre_normalized(0, 1, lmax1+lmax2+1)[0][:,:,0]
        for im1, m1 in enumerate(range(-mmax1, mmax1 + 1)):
            for im2, m2 in enumerate(range(-mmax2, mmax2 + 1)):
                for il1, l1 in enumerate(range(max(1, abs(m1)), lmax1 + 1)):
                    for il2, l2 in enumerate(range(max(1, abs(m2)), lmax2 + 1)):
                        for ild, ld in enumerate(range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1)): 
                            
                            a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)
                            leg =  legendre_2D[ld][abs(m1 - m2)]
                           
                            a5leg = a5*leg
                            b5leg = b5*leg
                            #ab5leg_key.append([l1, l2, m1, m2, ld])
                            a5leg_array.append(a5leg)
                            b5leg_array.append(b5leg)
                          
    
        return a5leg_array, b5leg_array


try:
    from smuthi.utility.cython.cython_speedups  import _ab5_coefficient_hash_table
    @memo.Memoize
    def ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2):
        return _ab5_coefficient_hash_table(int(lmax1), int(lmax2), int(mmax1), int(mmax2))  
except:
    @memo.Memoize
    def ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2):
        r"""Creates a hash table of the elements
        :math:`a5(l_1,m_1,l_2,m_2,l_d)` and :math:`a5(l_1,m_1,l_2,m_2,l_d)`
        found in appendix B of [Egel 2018 diss],where a5 and b5 are the coefficients used in the evaluation of the SVWF translation
        operator. This hash table is usefull in reducing computation time when calculating the coupling between two particles
        that exist in the same layer but do not have the same z coordinate.


        Args:
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle

    
        Returns:
            a5_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`
            b5_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`          

        """        
        a5_array = []
        b5_array = []
        #ab5_key = []
        for im1, m1 in enumerate(range(-mmax1, mmax1 + 1)):
            for im2, m2 in enumerate(range(-mmax2, mmax2 + 1)):
                for il1, l1 in enumerate(range(max(1, abs(m1)), lmax1 + 1)):
                    for il2, l2 in enumerate(range(max(1, abs(m2)), lmax2 + 1)):
                        for ild, ld in enumerate(range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1)): 
                            a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)
                            #ab5_key.append([l1, l2, m1, m2, ld])
                            a5_array.append(a5)
                            b5_array.append(b5)
                          
        return a5_array, b5_array


try:
    from smuthi.utility.cython.cython_speedups import direct_coupling_block_2D_from_hash_table 
except:
    def direct_coupling_block_2D_from_hash_table(blocksize1,blocksize2,
                                        w,sph,k,dx, dy, dz,
                                        lmax1, lmax2, mmax1, mmax2,
                                        a5leg_hash_table,b5leg_hash_table,
                                        threaded = False):

        r"""Subroutine to calculate the direct coupling between two particles 
            that are in the same layer and have the same z coordinate. This 
            subroutine is called internally by direct_coupling_block. 
            If Cython is enabled then this subroutine is Cython accelerated.
            Otherwise, the Python equivalent subroutine is used.


        Args:
            blocksize1 (int):           Number of columns in the direct coupling block
            blocksize2 (int):           Number of rows in the direct coupling block
            w  (ndarray):               Zero initialized direct coupling matrix block as numpy array
            sph  (ndarray):             Zero initialized Hankel function matrix as numpy array
            k (complex):                Wavenumber in the shared media
            dx (float):                 x-coordinate of the distance between the two particles
            dy  (float):                y-coordinate of the distance between the two particles
            dz  (float):                z-coordinate of the distance between the two particles
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle
            a5leg_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/phi` or :math:`kd`
            b5leg_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/phi` or :math:`kd`          
            threaded (bool):          Flag to enable multithreading (valid only for Cython where the Gil can be released). Currently hard coded to False.  
            
        Returns:
            w  (ndarray):          Direct coupling matrix block as numpy array.
        """


        # calculate distance vector
        d = np.sqrt(dx**2 + dy**2 + dz**2)
        phi = np.arctan2(dy, dx)
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = sma.spherical_hankel(int(n), complex(k * d))
        
        # the particle coupling operator is the transpose of the SVWF translation operator
        # therefore, (l1,m1) and (l2,m2) are interchanged:
        idx_ab5leg = 0
        for m1 in range(-mmax1, mmax1 + 1):
            for m2 in range(-mmax2, mmax2 + 1):
                eimph_val = np.exp(1j * (m2 - m1) * phi)
                for l1 in range(max(1, abs(m1)), lmax1 + 1):
                    for l2 in range(max(1, abs(m2)), lmax2 + 1):
                        A, B = complex(0), complex(0)
                        for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0

                            A += a5leg_hash_table[idx_ab5leg]*sph[ld] 
                            B += b5leg_hash_table[idx_ab5leg]*sph[ld]
                            idx_ab5leg += 1
                            
                        A, B = eimph_val * A, eimph_val * B
                        for tau1 in range(2):
                            n1 = flds.multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                            for tau2 in range(2):
                                n2 = flds.multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                                if tau1 == tau2:
                                    w[n1, n2] = A
                                else:
                                    w[n1, n2] = B
    
        return w
    

try:
    from smuthi.utility.cython.cython_speedups import direct_coupling_block_3D_from_hash_table
except:  
    def direct_coupling_block_3D_from_hash_table(blocksize1,blocksize2,
                                        w,sph,k,dx, dy, dz,
                                        lmax1, lmax2, mmax1, mmax2,
                                        a5_hash_table, b5_hash_table,
                                        threaded = False):
        
        r"""Subroutine to calculate the direct coupling between two particles 
            that are in the same layer and do not have the same z coordinate. This 
            subroutine is called internally by direct_coupling_block. 
            If Cython is enabled then this subroutine is Cython accelerated.
            Otherwise, the Python equivalent subroutine is used.


        Args:
            blocksize1 (int):           Number of columns in the direct coupling block
            blocksize2 (int):           Number of rows in the direct coupling block
            w  (ndarray):               Zero initialized direct coupling matrix block as numpy array
            sph  (ndarray):             Zero initialized Hankel function matrix as numpy array
            k (complex):                Wavenumber in the shared media
            dx (float):                 x-coordinate of the distance between the two particles
            dy  (float):                y-coordinate of the distance between the two particles
            dz  (float):                z-coordinate of the distance between the two particles
            lmax1 (int):           Largest polar quantum number of the recieving particle
            lmax2  (int):          Largest polar quantum number  of the emitting particle
            mmax1  (int):          Largest azimuthal quantum number of the recieving particle
            mmax2  (int):          Largest azimuthal quantum number of the emitting particle
            a5_array (ndarray):          Hash table of the elements in :math:`A` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`
            b5_array (ndarray):          Hash table of the elements in :math:`B` not dependent on :math:`/theta`, :math:`/phi` or :math:`kd`          
            threaded (bool):          Flag to enable multithreading (valid only for Cython where the Gil can be released). Currently hard coded to False.  
            
        Returns:
            w  (ndarray):          Direct coupling matrix block as numpy array.
        """        

        d = np.sqrt(dx**2 + dy**2 + dz**2)
        cos_theta = dz / d
        sin_theta = np.sqrt(dx**2 + dy**2) / d
        phi = np.arctan2(dy, dx)
    
        # spherical functions
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = sma.spherical_hankel(int(n), complex(k * d))
        legendre = sma.legendre_normalized(cos_theta, sin_theta, lmax1 + lmax2)[0][:,:,0]
        
        # the particle coupling operator is the transpose of the SVWF translation operator
        # therefore, (l1,m1) and (l2,m2) are interchanged:
        idx_ab5 = 0
        for m1 in range(-mmax1, mmax1 + 1):
            for m2 in range(-mmax2, mmax2 + 1):
                eimph_val = np.exp(1j * (m2 - m1) * phi)
                for l1 in range(max(1, abs(m1)), lmax1 + 1):
                    for l2 in range(max(1, abs(m2)), lmax2 + 1):
                        A, B = complex(0), complex(0)
                        for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                            A += a5_hash_table[idx_ab5]*sph[ld]*legendre[ld][abs(m1 - m2)] 
                            B += b5_hash_table[idx_ab5]*sph[ld]*legendre[ld][abs(m1 - m2)] 
                            idx_ab5 += 1
                            
                        A, B = eimph_val * A, eimph_val * B
                        for tau1 in range(2):
                            n1 = flds.multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                            for tau2 in range(2):
                                n2 = flds.multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                                if tau1 == tau2:
                                    w[n1, n2] = A
                                else:
                                    w[n1, n2] = B
    
        return w



###############################################################################
#                  PVWF coupling - experimental!                              #                               
###############################################################################
"""The following code section contains functions to compute the particle 
coupling via a PVWF expansion. This allows in principle to treat particles
with intersecting circumscribing spheres, see 
Theobald et al.: "Plane-wave coupling formalism for T-matrix simulations of 
light scattering by nonspherical particles", Phys Rev A, 2018"""


def spheroids_closest_points(ab_halfaxis1, c_halfaxis1, center1, orientation1, ab_halfaxis2, c_halfaxis2, center2, 
                             orientation2):
    """ Computation of the two closest points of two adjacent spheroids.
    For details, see: Stephen B. Pope, Algorithms for Ellipsoids, Sibley School of Mechanical & Aerospace Engineering, 
    Cornell University, Ithaca, New York, February 2008
    
    Args:
        ab_halfaxis1 (float):        Half axis orthogonal to symmetry axis of spheroid 1
        c_halfaxis1 (float):         Half axis parallel to symmetry axis of spheroid 1
        center1 (numpy.array):       Center coordinates of spheroid 1
        orientation1 (numpy.array):  Orientation angles of spheroid 1
        ab_halfaxis2 (float):        Half axis orthogonal to symmetry axis of spheroid 2
        c_halfaxis2 (float):         Half axis parallel to symmetry axis of spheroid 2
        center2 (numpy.array):       Center coordinates of spheroid 2
        orientation2 (numpy.array):  Orientation angles of spheroid 2
        
    Retruns:
        Tuple containing:
          - closest point on first particle (numpy.array)
          - closest point on second particle (numpy.array)
          - first rotation Euler angle alpha (float)
          - second rotation Euler angle beta (float)
    """
    
    def rotation_matrix(ang):
        rot_mat = (np.array([[np.cos(ang[0]) * np.cos(ang[1]), -np.sin(ang[0]), np.cos(ang[0]) * np.sin(ang[1])],
                             [np.sin(ang[0]) * np.cos(ang[1]), np.cos(ang[0]), np.sin(ang[0]) * np.sin(ang[1])],
                             [-np.sin(ang[1]), 0, np.cos(ang[1])]]))
        return rot_mat
    
    rot_matrix_1 = rotation_matrix(orientation1)
    rot_matrix_2 = rotation_matrix(orientation2)
        
    a1, a2 = ab_halfaxis1, ab_halfaxis2
    c1, c2 = c_halfaxis1, c_halfaxis2
    ctr1, ctr2 = np.array(center1), np.array(center2)
    
    eigenvalue_matrix_1 = np.array([[1 / a1 ** 2, 0, 0], [0, 1 / a1 ** 2, 0], [0, 0, 1 / c1 ** 2]])
    eigenvalue_matrix_2 = np.array([[1 / a2 ** 2, 0, 0], [0, 1 / a2 ** 2, 0], [0, 0, 1 / c2 ** 2]])
    
    E1 = np.dot(rot_matrix_1, np.dot(eigenvalue_matrix_1, np.transpose(rot_matrix_1)))
    E2 = np.dot(rot_matrix_2, np.dot(eigenvalue_matrix_2, np.transpose(rot_matrix_2)))
    S = np.matrix.getH(np.linalg.cholesky(E1))
    
    # transformation of spheroid E1 into the unit-sphere with its center at origin / same transformation on E2
    # E1_prime = np.dot(np.transpose(np.linalg.inv(S)), np.dot(E1, np.linalg.inv(S)))
    # ctr1_prime = np.zeros_like(ctr1)
    E2_prime = np.dot(np.transpose(np.linalg.inv(S)), np.dot(E2, np.linalg.inv(S)))
    ctr2_prime = -(np.dot(S, (ctr1 - ctr2)))  
    E2_prime_L = np.linalg.cholesky(E2_prime)
        
    H = np.dot(np.linalg.inv(E2_prime_L), np.transpose(np.linalg.inv(E2_prime_L)))
    p = np.array([0, 0, 0])
    f = np.dot(np.transpose(ctr2_prime - p), np.transpose(np.linalg.inv(E2_prime_L)))
    
    def minimization_fun(y_vec):
        fun = 0.5 * np.dot(np.dot(np.transpose(y_vec), H), y_vec) + np.dot(f, y_vec)
        return fun

    def constraint_fun(x):
        eq_constraint = (x[0] ** 2 + x[1] ** 2 + x[2] ** 2) ** 0.5 - 1
        return eq_constraint
    
    bnds = ((-1, 1), (-1, 1), (-1, 1))
    length_constraints = {'type' : 'eq', 'fun' : constraint_fun}
    
    flag = False
    while not flag:
        x0 = -1 + np.dot((1 + 1), np.random.rand(3))
        optimization_result = scipy.optimize.minimize(minimization_fun, x0, method='SLSQP', bounds=bnds,
                                                      constraints=length_constraints, tol=None, callback=None, options=None)
        x_vec = np.transpose(np.dot(np.transpose(np.linalg.inv(E2_prime_L)), optimization_result['x'])
                             + np.transpose(ctr2_prime))
        if optimization_result['success'] == True:
            if np.linalg.norm(x_vec) <= 1:
                raise ValueError("particle error: particles intersect")
            elif np.linalg.norm(x_vec) < np.linalg.norm(ctr2_prime):
                flag = True
            else:
                print('wrong minimum ...')
        else:
            print('No minimum found ...')

    p2_prime = x_vec
    p2 = np.dot(np.linalg.inv(S), p2_prime) + ctr1
    
    E1_L = np.linalg.cholesky(E1)
    H = np.dot(np.linalg.inv(E1_L), np.transpose(np.linalg.inv(E1_L)))
    p = p2
    f = np.dot(np.transpose(ctr1 - p), np.transpose(np.linalg.inv(E1_L)))
       
    flag = False
    while not flag:
        x0 = -1 + np.dot((1 + 1), np.random.rand(3))
        optimization_result2 = scipy.optimize.minimize(minimization_fun, x0, method='SLSQP', bounds=bnds,
                                                      constraints=length_constraints, tol=None, callback=None, options=None)
        p1 = np.transpose(np.dot(np.transpose(np.linalg.inv(E1_L)), optimization_result2['x']) + np.transpose(ctr1))
        if optimization_result2['success'] == True:
            if np.linalg.norm(p1 - p) < np.linalg.norm(ctr1 - p):
                flag = True
            else:
                print('wrong minimum ...')
        else:
            print('No minimum found ...')
    
    p1p2 = p2 - p1
    azimuth = np.arctan2(p1p2[1], p1p2[0])
    elevation = np.arctan2(p1p2[2], (p1p2[0] ** 2 + p1p2[1] ** 2) ** 0.5)

    if p1p2[2] < 0:
        beta = (np.pi / 2) + elevation
    else:
        beta = (-np.pi / 2) + elevation
    alpha = -azimuth
              
    return p1, p2, alpha, beta


def get_separating_plane(receiving_particle, emitting_particle):
    """Estimate the orientation of a plane that separates two particles.
    Such a plane always exists for convex particles.

    For some special cases (pairs of spheroids, pairs of non-rotated cylinders)
    a separating plane is constructed.

    For all other cases, the orientation is chosen along the center-to-center
    vector. Note that this choice is not guaranteed
    to yield a correct PVWF coupling result.

        Args:
        receiving_particle (smuthi.particles.Particle):        Receiving particle
        emitting_particle (smuthi.particles.Particle):         Emitting particle

    Retruns:
        Tuple containing Euler angles for the active rotation into a frame such
        that the emitting particle is above some z-plane and the receiving particle
        is below that plane:
          - first rotation Euler angle alpha (float)
          - second rotation Euler angle beta (float)
    """

    relpos = np.array(emitting_particle.position) - np.array(receiving_particle.position)

    if type(receiving_particle).__name__ == 'Spheroid' \
            and type(emitting_particle).__name__ == 'Spheroid':
        _, _, alpha, beta = spheroids_closest_points(
            emitting_particle.semi_axis_a, emitting_particle.semi_axis_c,
            emitting_particle.position, emitting_particle.euler_angles,
            receiving_particle.semi_axis_a, receiving_particle.semi_axis_c,
            receiving_particle.position, receiving_particle.euler_angles)
        return alpha, beta

    elif (type(receiving_particle).__name__ == 'FiniteCylinder'
            and type(emitting_particle).__name__ == 'FiniteCylinder'
            and np.linalg.norm(receiving_particle.euler_angles) == 0
            and np.linalg.norm(emitting_particle.euler_angles) == 0):
        in_plane_distance2 = relpos[0]**2 + relpos[1]**2
        sum_radii2 = (emitting_particle.cylinder_radius + receiving_particle.cylinder_radius)**2
        if in_plane_distance2 > sum_radii2:
            alpha = - np.arctan2(relpos[1], relpos[0])
            beta = - np.pi / 2
            return alpha, beta
        if relpos[2] > (emitting_particle.cylinder_height + receiving_particle.cylinder_height) / 2.0:
            return 0, 0
        if -relpos[2] > (emitting_particle.cylinder_height + receiving_particle.cylinder_height) / 2.0:
            return 0, np.pi

    warnings.warn("Unable to estimate suitable separating plane. Using center-to-center vector for PVWF coupling.")
    alpha = - np.arctan2(relpos[1], relpos[0])
    beta = - np.arccos(relpos[2] / np.linalg.norm(relpos))
    return alpha, beta


def direct_coupling_block_pvwf_mediated(vacuum_wavelength, receiving_particle, emitting_particle, layer_system,
                                        k_parallel, alpha=None, beta=None):
    """Direct particle coupling matrix :math:`W` for two particles (via plane vector wave functions).
    For details, see:
    Dominik Theobald et al., Phys. Rev. A 96, 033822, DOI: 10.1103/PhysRevA.96.033822 or arXiv:1708.04808

    The plane wave coupling is performed in a rotated coordinate system,
    which must be chosen such that both particles can be separated by a plane
    that is parallel to the xy-plane (such that the emitting particle is
    entirely above that plane and the receiving particle is entirely below that
    plane).

    Two angles (alpha and beta) are required to specify the active rotation into
    that coordinate system, i.e., the rotation which rotates the particle
    locations such that the abovementioned condition is fulfilled.

    If the angle arguments are omitted, Smuthi tries to estimate a suitable
    separating plane.

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place
        k_parallel (numpy.array):                           In-plane wavenumber for plane wave expansion
        alpha (float):                                      First Euler angle, rotation around z-axis, in rad
        beta (float):                                       Second Euler angle, rotation around y'-axis in rad

    Returns:
        Direct coupling matrix block (numpy array).
    """    
    lmax1 = receiving_particle.l_max
    mmax1 = receiving_particle.m_max
    assert lmax1 == mmax1, 'PVWF coupling requires lmax == mmax for each particle.'
    lmax2 = emitting_particle.l_max
    mmax2 = emitting_particle.m_max
    assert lmax2 == mmax2, 'PVWF coupling requires lmax == mmax for each particle.'
    lmax = max([lmax1, lmax2])
    blocksize1 = flds.blocksize(lmax1, mmax1)
    blocksize2 = flds.blocksize(lmax2, mmax2)

    # initialize result
    w = np.zeros((blocksize1, blocksize2), dtype=complex)

    # Check if particles are in the same layer.
    rS1 = receiving_particle.position
    rS2 = emitting_particle.position
    iS1 = layer_system.layer_number(rS1[2])
    iS2 = layer_system.layer_number(rS2[2])
    if (iS1 != iS2) or emitting_particle == receiving_particle:
        return w

    n_medium = layer_system.refractive_indices[layer_system.layer_number(receiving_particle.position[2])]
    
    if alpha is None or beta is None:
        alpha, beta = get_separating_plane(receiving_particle, emitting_particle)

    # positions
    r1 = np.array(receiving_particle.position)
    r2 = np.array(emitting_particle.position)
    r21_lab = r1 - r2  # laboratory coordinate system
    
    # distance vector in rotated coordinate system
    r21_rot = np.dot(np.dot([[np.cos(beta), 0, np.sin(beta)], [0, 1, 0], [- np.sin(beta), 0, np.cos(beta)]],
                           [[np.cos(alpha), - np.sin(alpha), 0], [np.sin(alpha), np.cos(alpha), 0], [0, 0, 1]]), 
                    r21_lab)
    rho21 = (r21_rot[0] ** 2 + r21_rot[1] ** 2) ** 0.5 
    phi21 = np.arctan2(r21_rot[1], r21_rot[0])
    z21 = r21_rot[2]
    
    # wavenumbers
    omega = flds.angular_frequency(vacuum_wavelength)
    k = omega * n_medium
    kz = flds.k_z(k_parallel=k_parallel, vacuum_wavelength=vacuum_wavelength, refractive_index=n_medium)
    if z21 < 0:
        kz_var = -kz
    else:
        kz_var = kz
        
    # Bessel lookup 
    bessel_list = []
    for dm in range(mmax1 + mmax2 + 1):
        bessel_list.append(scipy.special.jn(dm, k_parallel * rho21))
    
    # legendre function lookups
    ct = kz_var / k
    st = k_parallel / k
    _, pilm_list, taulm_list = sma.legendre_normalized(ct, st, lmax)
    
    # prefactor
    const_arr = k_parallel / (kz * k) * np.exp(1j * (kz_var * z21))
                        
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            jmm_eimphi_bessel = 4 * 1j ** abs(m2 - m1) * np.exp(1j * phi21 * (m2 - m1)) * bessel_list[abs(m2 - m1)]
            prefactor = const_arr * jmm_eimphi_bessel
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    for tau1 in range(2):
                        n1 = flds.multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = flds.multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            for pol in range(2):
                                B_dag = trf.transformation_coefficients_vwf(tau1, l1, m1, pol, pilm_list=pilm_list,
                                                                        taulm_list=taulm_list, dagger=True)
                                B = trf.transformation_coefficients_vwf(tau2, l2, m2, pol, pilm_list=pilm_list,
                                                                            taulm_list=taulm_list, dagger=False)
                                integrand = prefactor * B * B_dag
                                w[n1, n2] += np.trapz(integrand, k_parallel) 
                                
    rot_mat_1 = trf.block_rotation_matrix_D_svwf(lmax1, mmax1, 0, beta, alpha)
    rot_mat_2 = trf.block_rotation_matrix_D_svwf(lmax2, mmax2, -alpha, -beta, 0)
    
    return np.dot(np.dot(np.transpose(rot_mat_1), w), np.transpose(rot_mat_2))



# -*- coding: utf-8 -*-
"""This module is needed for the installation of the package."""

from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.build_ext import build_ext

import os
import sys
import subprocess

from wheel.bdist_wheel import bdist_wheel
from numpy.distutils.core import setup
from numpy.distutils.core import Extension

import pkg_resources
import numpy as np

version = {}
with open("smuthi/version.py") as fp:
    exec(fp.read(), version)
__version__ = version['__version__']


def get_numpy_version():
    import numpy as np
    return np.__version__


class PrepareCommand(develop):
    def run(self):
        prepare_nfmds()
        

class CustomDevelopCommand(develop):
    def run(self):
        print("**********"
              +"\nrunning develop with python " + sys.version 
              + "and numpy " + get_numpy_version() 
              + "\n**********")
        prepare_nfmds()
        develop.run(self)


class CustomInstallCommand(install):
    def run(self):
        print("**********"
              +"\nrunning install with python " + sys.version 
              + "and numpy " + get_numpy_version() 
              + "\n**********")        
        prepare_nfmds()
        install.run(self)


class CustomBuildExtCommand(build_ext):
    def run(self):
        print("**********"
              +"\nrunning build_ext with python " + sys.version 
              + "and numpy " + get_numpy_version() 
              + "\n**********")    
        build_ext.run(self)


class CustomBdistWheelCommand(bdist_wheel):
    def run(self):
        print("**********"
              +"\nrunning bdist_wheel with python " + sys.version 
              + "and numpy " + get_numpy_version() 
              + "\n**********")    
        if sys.platform.startswith('win'):
            # Skip F2Py. Before "python setup.py bdist_wheel", you need to call in advance:
            # python setup.py prepare
            # python setup.py build_ext --inplace --compiler=mingw32 --fcompiler=gnu95 -f
            self.distribution.ext_modules = []
        bdist_wheel.run(self)


def prepare_nfmds():
    """If Windows: Call encoding converter to get Fortran sources with valid encoding"""
    if sys.platform.startswith('win'):
        currdir = os.getcwd()
        nfmds_sources_dirname = pkg_resources.resource_filename('smuthi.linearsystem.tmatrix.nfmds', 'NFM-DS')
        os.chdir(nfmds_sources_dirname + '/TMATSOURCES')
        with open("encoding_converter.py") as fp:
            exec(fp.read(), version)
        os.chdir(currdir)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_requirements():
    """Return a list of requirements, depending on the operating system."""
    requirements = ['numpy>=1.19.3',
                    'argparse',
                    'imageio',
                    'matplotlib',
                    'mpmath',
                    'numba',
                    'pyyaml',
                    'scipy',
                    'sympy',
                    'tqdm',
                    'h5py',
                    'pycparser',
                    'psutil']
    
    if sys.platform.startswith('win'):
        # this package offers windows binaries:
        requirements.append('pywigxjpf-win')
    else:
        requirements.append('pywigxjpf')
        
    return requirements


def get_extensions(static=True):
    """Depending on the platform, pick suitable Extension object. This is important such that if
    MinGW is used, the DLLs are statically linked (otherwise the user cannot use the binary if he 
    doesn't have MinGW on his computer, too)."""    
    if os.environ.get('READTHEDOCS'):
        return []    
    f2py_options = ['only:', 'tlay','taxsym','tnonaxsym', ':']
    static_link_args = []    
    if sys.platform.startswith('win'):
        if static:
            static_link_args = ["-static", "-static-libgfortran", "-static-libgcc"]            
        nfmds_path = 'smuthi/linearsystem/tmatrix/nfmds/NFM-DS/TMATSOURCES/win/TAXSYM_SMUTHI.f90'
    else:
        nfmds_path = 'smuthi/linearsystem/tmatrix/nfmds/NFM-DS/TMATSOURCES/TAXSYM_SMUTHI.f90'    
    extensions = [Extension('smuthi.linearsystem.tmatrix.nfmds.nfmds', [nfmds_path],
                            extra_link_args=static_link_args, f2py_options=f2py_options),
                  Extension("smuthi.utility.cython.cython_speedups", 
                            ["smuthi/utility/cython/cython_speedups.c"],
                            extra_compile_args=['-fopenmp'],
                            extra_link_args=static_link_args+['-fopenmp'],
                            include_dirs=[np.get_include()])]    
    return extensions

setup(
    name="SMUTHI",
    version=__version__,
    author="Amos Egel",
    author_email="amos.egel@gmail.com",
    url='https://gitlab.com/AmosEgel/smuthi',
    description="Light scattering by multiple particles in thin-film systems",
    long_description=read('README.rst'),
    packages=['smuthi',
              'smuthi.fields',
              'smuthi.linearsystem',
              'smuthi.linearsystem.tmatrix',
              'smuthi.linearsystem.tmatrix.nfmds',
              'smuthi.linearsystem.particlecoupling',
              'smuthi.postprocessing',
              'smuthi.periodicboundaries',
              'smuthi.utility',
              'smuthi.utility.cython'],
    ext_modules=get_extensions(True),
    cmdclass={'prepare': PrepareCommand,
              'compile': build_ext,
              'install': CustomInstallCommand,
              'develop': CustomDevelopCommand,
              'bdist_wheel': CustomBdistWheelCommand},
    package_data={'smuthi.linearsystem.tmatrix.nfmds': ['NFM-DS/*.txt', 'NFM-DS/TMATSOURCES/*.f90', 'NFM-DS/TMATFILES/*',
                                                        'NFM-DS/INPUTFILES/*.dat', 'NFM-DS/OUTPUTFILES/*','nfmds*'],
                  'smuthi.utility.cython': ['cython_speedups*']},
    include_package_data=True,                  
    install_requires=get_requirements(),
    extras_require={'cuda':  ['PyCuda']},
    license='MIT',
)



major = 2
minor = 1
micro = 2

pre_release = None
post_release = None
dev_release = None

__version__ = '{}'.format(major)

if minor is not None:
    __version__ += '.{}'.format(minor)

if micro is not None:
    __version__ += '.{}'.format(micro)

if pre_release is not None:
    __version__ += '{}'.format(pre_release)

if post_release is not None:
    __version__ += '.post{}'.format(post_release)

if dev_release is not None:
    __version__ += '.dev{}'.format(dev_release)



import numpy as np
import smuthi.fields as flds
from numba import jit


@jit(nopython=True, cache=True)
def single_index_to_multi_nfmds(index, Nrank, Mrank):
    """Converts single index to (tau,l,m) tuple in NFMDS convention.

    Args:
        index (int):    single index in NFMDS convention    
        Nrank (int):    NFMDS Nrank parameter
        Mrank (int):    NFMDS Mrank parameter       

    Returns:
        tau (int):      SVWF polarization (0 for spherical TE, 1 for spherical TM)
        l (int):        SVWF degree
        m (int):        SVWF order    
    """      
    
    index = index + 1
    nmax = Nrank + Mrank * (2 * Nrank - Mrank + 1)
    if index > nmax:
        tau = 1
        index -= nmax
    else:
        tau = 0

    for m in range(Mrank + 1):
        N0 = (Nrank + (abs(m + 1) - 1) * (2 * Nrank - abs(m + 1) + 2))
        if index <= N0:
            break
    if m > 0:
        N0 = (Nrank + (m - 1) * (2 * Nrank - m + 2))
        index -= N0
    if index > (Nrank - abs(m) + 1):
        index -= (Nrank - abs(m) + 1)
        m = -m
    l = index + (abs(m) - 1) * (abs(m) > 0)
    return tau, l, m


def multi_index_to_single_nfmds(tau, l, m, Nrank, Mrank):
    """Converts a (tau,l,m) index to single index in NFMDS convention.

    Args:
        tau (int):      SVWF polarization (0 for spherical TE, 1 for spherical TM)
        l (int):        SVWF degree
        m (int):        SVWF order        
        Nrank (int):    NFMDS Nrank parameter
        Mrank (int):    NFMDS Mrank parameter              

    Returns:
        index (int):    single index in NFMDS convention 
    """      
    nmax = Nrank + Mrank * (2 * Nrank - Mrank + 1)
    if m == 0:
        index = l
    else:
        N0 = Nrank + (abs(m) - 1) * (2 * Nrank - abs(m) + 2)

        if m < 0:
            N0 = N0 + Nrank - abs(m) + 1
        index = N0 + l - abs(m) - 1
    if tau:
        index = index + nmax
    return index - 1  # cause python numbers from 0 and fortran from 1


@jit(nopython=True, cache=True)
def nfmds_to_smuthi_matrix(T, Nrank=None, Mrank=None, l_max=None, m_max=None):
    """Converts a T-matrix obtained with NFMDS to SMUTHI compatible format.

    Args:
        T (array):      T-matrix in NFMDS convention
        Nrank (int):    NFMDS Nrank parameter
        Mrank (int):    NFMDS Mrank parameter
        l_max (int):    Maximal multipole degree used for the spherical wave expansion of incoming and
                        scattered field
        m_max (int):    Maximal multipole order used for the spherical wave expansion of incoming and
                        scattered field
       
    Returns:
        Tsm (array):    T-matrix in SMUTHI convention
    """

    if Nrank is None:
        nmax = T.shape[0] / 2
        Nrank = int(-1 + np.sqrt(1 + nmax))
        Mrank = int(Nrank)

    if Mrank is None:
        Mrank = Nrank

    if l_max is None:
        l_max = Nrank

    if m_max is None:
        m_max = l_max

    Tsm = np.zeros((flds.blocksize(l_max, m_max), flds.blocksize(l_max, m_max)), dtype=np.complex128)
    num_cols = range(T.shape[0])
    num_rows = range(T.shape[1])
    for ii in num_cols:
        tau, l_ii, m_ii = single_index_to_multi_nfmds(ii, Nrank, Mrank)
        if l_ii > l_max or abs(m_ii) > m_max:
            continue
        ii_sm = flds.multi_to_single_index(tau, l_ii, m_ii, l_max, m_max)
        for jj in num_rows:
            tau, l_jj, m_jj = single_index_to_multi_nfmds(jj, Nrank, Mrank)
            if l_jj > l_max or abs(m_jj) > m_max:
                continue
            jj_sm = flds.multi_to_single_index(tau, l_jj, m_jj, l_max, m_max)
            Tsm[ii_sm, jj_sm] = T[ii, jj]
    return Tsm


@jit(nopython=True, cache=True)
def python_to_smuthi_matrix(T, Nrank, Mrank=None, l_max=None, m_max=None):
    """Converts a T-matrix obtained with Alan's code to SMUTHI compatible format.

    Args:
        T (array):      T-matrix in NFMDS convention
        Nrank (int):    Alan's lmax parameter
        Mrank (int):    Alan's lmax parameter
        l_max (int):    Maximal multipole degree used for the spherical wave expansion of incoming and
                        scattered field
        m_max (int):    Maximal multipole order used for the spherical wave expansion of incoming and
                        scattered field

    Returns:
        Tsm (array):    T-matrix in SMUTHI convention
    """

    if Mrank is None:
        Mrank = Nrank

    if l_max is None:
        l_max = Nrank

    if m_max is None:
        m_max = l_max

    Tsm = np.zeros((flds.blocksize(l_max, m_max), flds.blocksize(l_max, m_max)), dtype=np.complex128)

    mlim = min(m_max, Mrank)
    llim = min(l_max, Nrank)

    for tau1 in range(2):
        for m1 in range(-mlim, mlim + 1):
            for l1 in range(max(1, abs(m1)), llim + 1):
                n1_py = flds.multi_to_single_index(tau1, l1, m1, Nrank, Mrank)
                n1_sm = flds.multi_to_single_index(tau1, l1, m1, l_max, m_max)
                for tau2 in range(2):
                    for m2 in range(-mlim, mlim + 1):
                        for l2 in range(max(1, abs(m2)), llim + 1):
                            n2_py = flds.multi_to_single_index(tau2, l2, m2, Nrank, Mrank)
                            n2_sm = flds.multi_to_single_index(tau2, l2, m2, l_max, m_max)
                            Tsm[n1_sm, n2_sm] = T[n1_py, n2_py]

    return Tsm



"""This module contains functionality to prepare lookups for the particle 
coupling coefficients, which allows to efficiently treat large numbers of 
particles. """

import sys
from tqdm import tqdm
import numpy as np
import scipy.special
import smuthi.fields
import smuthi.utility.math as sf
import smuthi.utility.cuda as cu
import smuthi.fields.transformations as trf
import smuthi.layers as lay
import smuthi.linearsystem.particlecoupling.prepare_lookup_cuda as cusrc

def volumetric_coupling_lookup_table(vacuum_wavelength, particle_list, layer_system, k_parallel='default', 
                                     resolution=None):
    """Prepare Sommerfeld integral lookup table to allow for a fast calculation of the coupling matrix by interpolation.
    This function is called when not all particles are on the same z-position.
    
    Args:
        vacuum_wavelength (float):  Vacuum wavelength in length units
        particle_list (list):       List of particle objects
        layer_system (smuthi.layers.LayerSystem):    Stratified medium
        k_parallel (numpy.ndarray or str):           In-plane wavenumber for Sommerfeld integrals.
                                                     If 'default', smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float): Spatial resolution of lookup table in length units. (default: vacuum_wavelength / 100)        
                            Smaller means more accurate but higher memory footprint 
                            
    Returns:
        (tuple): tuple containing:

            w_pl (ndarray):  Coupling lookup for z1 + z2, indices are [rho, z, n1, n2]. Includes layer mediated coupling.
            w_mn (ndarray):  Coupling lookup for z1 + z2, indices are [rho, z, n1, n2]. Includes layer mediated and 
                             direct coupling.
            rho_array (ndarray):    Values for the radial distance considered for the lookup (starting from negative 
                                    numbers to allow for simpler cubic interpolation without distinction of cases 
                                    for lookup edges
            sz_array (ndarray):     Values for the sum of z-coordinates (z1 + z2) considered for the lookup
            dz_array (ndarray):     Values for the difference of z-coordinates (z1 - z2) considered for the lookup 
    """
    sys.stdout.write('Prepare 3D particle coupling lookup:\n')
    sys.stdout.flush()

    if resolution is None:
        resolution = vacuum_wavelength / 100
        sys.stdout.write('Setting lookup resolution to %f\n'%resolution)
        sys.stdout.flush()

    l_max = max([particle.l_max for particle in particle_list])
    m_max = max([particle.m_max for particle in particle_list])
    blocksize = smuthi.fields.blocksize(l_max, m_max)
    
    particle_x_array = np.array([particle.position[0] for particle in particle_list])
    particle_y_array = np.array([particle.position[1] for particle in particle_list])
    particle_z_array = np.array([particle.position[2] for particle in particle_list])
    particle_rho_array = np.sqrt((particle_x_array[:, None] - particle_x_array[None, :]) ** 2 
                                 + (particle_y_array[:, None] - particle_y_array[None, :]) ** 2)
    
    dz_min = particle_z_array.min() - particle_z_array.max()
    dz_max = particle_z_array.max() - particle_z_array.min()
    sz_min = 2 * particle_z_array.min()
    sz_max = 2 * particle_z_array.max()
    
    rho_array = np.arange(- 3 * resolution, particle_rho_array.max() + 3 * resolution, resolution)
    sz_array = np.arange(sz_min - 3 * resolution, sz_max + 3 * resolution, resolution)
    dz_array = np.arange(dz_min - 3 * resolution, dz_max + 3 * resolution, resolution)
    
    len_rho = len(rho_array)
    len_sz = len(sz_array)
    len_dz = len(dz_array)
    assert len_sz == len_dz
    
    i_s = layer_system.layer_number(particle_list[0].position[2])
    k_is = layer_system.wavenumber(i_s, vacuum_wavelength)
    z_is = layer_system.reference_z(i_s)

    wr_pl = np.zeros((len_rho, len_dz, blocksize, blocksize), dtype=np.complex64)
    wr_mn = np.zeros((len_rho, len_dz, blocksize, blocksize), dtype=np.complex64)

    # direct -----------------------------------------------------------------------------------------------------------
    w = np.zeros((len_rho, len_dz, blocksize, blocksize), dtype=np.complex64)
    sys.stdout.write('Lookup table memory footprint: ' + size_format(2 * w.nbytes) + '\n')
    sys.stdout.flush()

    r_array = np.sqrt(dz_array[None, :]**2 + rho_array[:, None]**2)
    r_array[r_array==0] = 1e-20
    ct = dz_array[None, :] / r_array
    st = rho_array[:, None] / r_array
    legendre, _, _ = sf.legendre_normalized(ct, st, 2 * l_max)
    
    bessel_h = []
    for dm in tqdm(range(2 * l_max + 1), desc='Spherical Hankel lookup   ', file=sys.stdout,
                   bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}'):
        bessel_h.append(sf.spherical_hankel(dm, k_is * r_array))
    
    pbar = tqdm(total=blocksize**2, 
                desc='Direct coupling           ', 
                file=sys.stdout,
                bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')
            
    for m1 in range(-m_max, m_max+1):
        for m2 in range(-m_max, m_max+1):
            for l1 in range(max(1, abs(m1)), l_max + 1):
                for l2 in range(max(1, abs(m2)), l_max + 1):
                    A = np.zeros((len_rho, len_dz), dtype=complex)
                    B = np.zeros((len_rho, len_dz), dtype=complex)
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                        a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)    # remember that w = A.T
                        A += a5 * bessel_h[ld] * legendre[ld, abs(m1 - m2)]   # remember that w = A.T
                        B += b5 * bessel_h[ld] * legendre[ld, abs(m1 - m2)]   # remember that w = A.T
                    for tau1 in range(2):
                        n1 = smuthi.fields.multi_to_single_index(tau1, l1, m1, l_max, m_max)
                        for tau2 in range(2):
                            n2 = smuthi.fields.multi_to_single_index(tau2, l2, m2, l_max, m_max)
                            if tau1 == tau2:
                                w[:, :, n1, n2] = A
                            else:
                                w[:, :, n1, n2] = B
                            pbar.update()
    pbar.close()

    # switch off direct coupling contribution near rho=0:
    w[rho_array < particle_rho_array[~np.eye(particle_rho_array.shape[0],dtype=bool)].min() / 2, :, :, :] = 0  


    # Do not calculate the layer mediated lookup if there is no meaningful layer system!
    if layer_system.is_degenerate() == False:
        # layer mediated ---------------------------------------------------------------------------------------------------
        sys.stdout.write('Layer mediated coupling   : ...')
        sys.stdout.flush()
        if type(k_parallel) == str and k_parallel == 'default':
            k_parallel = smuthi.fields.default_Sommerfeld_k_parallel_array
        kz_is = smuthi.fields.k_z(k_parallel=k_parallel, k=k_is)
        len_kp = len(k_parallel)
    
        # phase factors
        epljksz = np.exp(1j * kz_is[None, :] * (sz_array[:, None] - 2 * z_is))  # z, k
        emnjksz = np.exp(- 1j * kz_is[None, :] * (sz_array[:, None] - 2 * z_is))
        epljkdz = np.exp(1j * kz_is[None, :] * dz_array[:, None])
        emnjkdz = np.exp(- 1j * kz_is[None, :] * dz_array[:, None])
           
        # layer response
        L = np.zeros((2, 2, 2, len_kp), dtype=complex)  # pol, pl/mn1, pl/mn2, kp
        for pol in range(2):
            L[pol, :, :, :] = lay.layersystem_response_matrix(pol, layer_system.thicknesses,
                                                              layer_system.refractive_indices, k_parallel,
                                                              smuthi.fields.angular_frequency(vacuum_wavelength), i_s, i_s)
       
        # transformation coefficients
        B_dag = np.zeros((2, 2, blocksize, len_kp), dtype=complex)  # pol, pl/mn, n, kp
        B = np.zeros((2, 2, blocksize, len_kp), dtype=complex)  # pol, pl/mn, n, kp
        ct_k = kz_is / k_is
        st_k = k_parallel / k_is
        _, pilm_pl, taulm_pl = sf.legendre_normalized(ct_k, st_k, l_max)
        _, pilm_mn, taulm_mn = sf.legendre_normalized(-ct_k, st_k, l_max)
        
        m_list = [None for i in range(blocksize)]
        for tau in range(2):
            for m in range(-m_max, m_max + 1):
                for l in range(max(1, abs(m)), l_max + 1):
                    n = smuthi.fields.multi_to_single_index(tau, l, m, l_max, m_max)
                    m_list[n] = m
                    for pol in range(2):
                        B_dag[pol, 0, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                                        taulm_list=taulm_pl, dagger=True)
                        B_dag[pol, 1, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_mn,
                                                                                        taulm_list=taulm_mn, dagger=True)
                        B[pol, 0, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                                 taulm_list=taulm_pl, dagger=False)
                        B[pol, 1, n, :] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_mn,
                                                                                 taulm_list=taulm_mn, dagger=False)
        
        # pairs of (n1, n2), listed by abs(m1-m2)
        n1n2_combinations = [[] for dm in range(2*m_max+1)]
        for n1 in range(blocksize):
            m1 = m_list[n1]
            for n2 in range(blocksize):
                m2 = m_list[n2]
                n1n2_combinations[abs(m1-m2)].append((n1,n2))
                       

        
        dkp = np.diff(k_parallel)
        if cu.use_gpu:
            re_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.real))
            im_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.imag))
            kernel_source_code = cusrc.volume_lookup_assembly_code %(blocksize, len_rho, len_sz, len_kp)
            helper_function = cu.SourceModule(kernel_source_code).get_function("helper")
            cuda_blocksize = 128
            cuda_gridsize = (len_rho * len_sz + cuda_blocksize - 1) // cuda_blocksize
    
            re_dwr_d = cu.gpuarray.to_gpu(np.zeros((len_rho, len_sz), dtype=np.float32))
            im_dwr_d = cu.gpuarray.to_gpu(np.zeros((len_rho, len_sz), dtype=np.float32))
    
        pbar = tqdm(total=blocksize**2, 
                    desc='Layer mediated coupling   ', 
                    file=sys.stdout,
                    bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')
                
        for dm in range(2*m_max+1):
            bessel = scipy.special.jv(dm, (k_parallel[None,:]*rho_array[:,None]))
            besjac = bessel * (k_parallel / (kz_is * k_is))[None,:]
    
            for n1n2 in n1n2_combinations[dm]:
                n1 = n1n2[0]
                m1 = m_list[n1]
                n2 = n1n2[1]
                m2 = m_list[n2]
        
                belbee_pl = np.zeros((len_dz, len_kp), dtype=complex)
                belbee_mn = np.zeros((len_dz, len_kp), dtype=complex)
                for pol in range(2):
                    belbee_pl += ((L[pol, 0, 1, :] * B_dag[pol, 0, n1, :] * B[pol, 1, n2, :])[None, :] * epljksz
                                  + (L[pol, 1, 0, :] * B_dag[pol, 1, n1, :] * B[pol, 0, n2, :])[None, :] * emnjksz)
                    belbee_mn += ((L[pol, 0, 0, :] * B_dag[pol, 0, n1, :] * B[pol, 0, n2, :])[None, :] * epljkdz
                                  + (L[pol, 1, 1, :] * B_dag[pol, 1, n1, :] * B[pol, 1, n2, :])[None, :] * emnjkdz)
                
                if cu.use_gpu:
                    re_belbee_pl_d = cu.gpuarray.to_gpu(np.float32(belbee_pl[None, :, :].real))
                    im_belbee_pl_d = cu.gpuarray.to_gpu(np.float32(belbee_pl[None, :, :].imag))
                    re_belbee_mn_d = cu.gpuarray.to_gpu(np.float32(belbee_mn[None, :, :].real))
                    im_belbee_mn_d = cu.gpuarray.to_gpu(np.float32(belbee_mn[None, :, :].imag))
    
                    re_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac[:, None, :].real))
                    im_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac[:, None, :].imag))
                    helper_function(re_besjac_d.gpudata, im_besjac_d.gpudata, re_belbee_pl_d.gpudata,
                                    im_belbee_pl_d.gpudata, re_dkp_d.gpudata, im_dkp_d.gpudata, re_dwr_d.gpudata, 
                                    im_dwr_d.gpudata, block=(cuda_blocksize, 1, 1), grid=(cuda_gridsize, 1))
                    wr_pl[:, :, n1, n2] = 4 * (1j)**abs(m2 - m1) * (re_dwr_d.get() + 1j * im_dwr_d.get()) 
                    
                    helper_function(re_besjac_d.gpudata, im_besjac_d.gpudata, re_belbee_mn_d.gpudata, 
                                    im_belbee_mn_d.gpudata, re_dkp_d.gpudata, im_dkp_d.gpudata, re_dwr_d.gpudata, 
                                    im_dwr_d.gpudata, block=(cuda_blocksize, 1, 1), grid=(cuda_gridsize, 1))
                    wr_mn[:, :, n1, n2] = 4 * (1j)**abs(m2 - m1) * (re_dwr_d.get() + 1j * im_dwr_d.get()) 
                else:
                    integrand = besjac[:, None, :] * belbee_pl[None, :, :] 
                    wr_pl[:, :, n1, n2] = 2 * (1j)**abs(m2 - m1) * ((integrand[:, :, :-1] + integrand[:, :, 1:]) 
                                                                    * dkp[None, None, :]).sum(axis=-1)   # trapezoidal rule
                    
                    integrand = besjac[:, None, :] * belbee_mn[None, :, :]
                    wr_mn[:, :, n1, n2] = 2 * (1j)**abs(m2 - m1) * ((integrand[:, :, :-1] + integrand[:, :, 1:])
                                                                    * dkp[None, None, :]).sum(axis=-1)
                pbar.update()
        pbar.close()
    else:
        sys.stdout.write('Omitting calculations of layer mediated coupling since no difference in layers was detected. \n')
        sys.stdout.flush()
    
    return wr_pl, w + wr_mn, rho_array, sz_array, dz_array


def radial_coupling_lookup_table(vacuum_wavelength, particle_list, layer_system, k_parallel='default', resolution=None):
    """Prepare Sommerfeld integral lookup table to allow for a fast calculation of the coupling matrix by interpolation.
    This function is called when all particles are on the same z-position.
    
    Args:
        vacuum_wavelength (float):  Vacuum wavelength in length units
        particle_list (list):       List of particle objects
        layer_system (smuthi.layers.LayerSystem):    Stratified medium
        k_parallel (numpy.ndarray or str):           In-plane wavenumber for Sommerfeld integrals.
                                                     If 'default', smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float): Spatial resolution of lookup table in length units. (default: vacuum_wavelength / 100)       
                            Smaller means more accurate but higher memory footprint 
                            
    Returns:
        (tuple) tuple containing:
        
            lookup_table (ndarray):  Coupling lookup, indices are [rho, n1, n2].
            rho_array (ndarray):     Values for the radial distance considered for the lookup (starting from negative 
                                     numbers to allow for simpler cubic interpolation without distinction of cases 
                                     at rho=0)
    """

    sys.stdout.write('Prepare radial particle coupling lookup:\n')
    sys.stdout.flush()

    if resolution is None:
        resolution = vacuum_wavelength / 100
        sys.stdout.write('Setting lookup resolution to %f\n'%resolution)
        sys.stdout.flush()
   
    l_max = max([particle.l_max for particle in particle_list])
    m_max = max([particle.m_max for particle in particle_list])
    blocksize = smuthi.fields.blocksize(l_max, m_max)
    
    x_array = np.array([particle.position[0] for particle in particle_list])
    y_array = np.array([particle.position[1] for particle in particle_list])
    rho_array = np.sqrt((x_array[:, None] - x_array[None, :]) ** 2 + (y_array[:, None] - y_array[None, :]) ** 2)

    radial_distance_array = np.arange(- 3 * resolution, rho_array.max() + 3 * resolution, resolution)
    
    z = particle_list[0].position[2]
    i_s = layer_system.layer_number(z)
    k_is = layer_system.wavenumber(i_s, vacuum_wavelength)
    dz = z - layer_system.reference_z(i_s)
    
    len_rho = len(radial_distance_array)
        
    # direct -----------------------------------------------------------------------------------------------------------
    w = np.zeros((len_rho, blocksize, blocksize), dtype=np.complex64)
    sys.stdout.write('Memory footprint: ' + size_format(w.nbytes) + '\n')
    sys.stdout.flush()

    ct = np.array([0.0])
    st = np.array([1.0])
    bessel_h = []
    for n in range(2* l_max + 1):
        bessel_h.append(sf.spherical_hankel(n, k_is * radial_distance_array))
        bessel_h[-1][radial_distance_array <= 0] = np.nan
        
    legendre, _, _ = sf.legendre_normalized(ct, st, 2 * l_max)

    pbar = tqdm(total=blocksize**2, 
                desc='Direct coupling           ', 
                file=sys.stdout,
                bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')

    for m1 in range(-m_max, m_max+1):
        for m2 in range(-m_max, m_max+1):
            for l1 in range(max(1, abs(m1)), l_max + 1):
                for l2 in range(max(1, abs(m2)), l_max + 1):
                    A = np.zeros(len_rho, dtype=complex)
                    B = np.zeros(len_rho, dtype=complex)
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                        a5, b5 = trf.ab5_coefficients(l2, m2, l1, m1, ld)
                        A = A + a5 * bessel_h[ld] * legendre[ld, abs(m1 - m2)]
                        B = B + b5 * bessel_h[ld] * legendre[ld, abs(m1 - m2)]
                    for tau1 in range(2):
                        n1 = smuthi.fields.multi_to_single_index(tau1, l1, m1, l_max, m_max)
                        for tau2 in range(2):
                            n2 = smuthi.fields.multi_to_single_index(tau2, l2, m2, l_max, m_max)
                            if tau1 == tau2:
                                w[:, n1, n2] = A  # remember that w = A.T
                            else:
                                w[:, n1, n2] = B
                            pbar.update()
    pbar.close()
    close_to_zero = radial_distance_array < rho_array[~np.eye(rho_array.shape[0],dtype=bool)].min() / 2 
    w[close_to_zero, :, :] = 0  # switch off direct coupling contribution near rho=0


    # Only calculate the layer mediated lookup if there is a meaningful layer system!
    if layer_system.is_degenerate() != True:
        # layer mediated ---------------------------------------------------------------------------------------------------
        sys.stdout.write('Layer mediated coupling   : ...')
        sys.stdout.flush()
        
        if type(k_parallel) == str and k_parallel == 'default':
            k_parallel = smuthi.fields.default_Sommerfeld_k_parallel_array
    
        kz_is = smuthi.fields.k_z(k_parallel=k_parallel, k=k_is)
        len_kp = len(k_parallel)
    
        # phase factors
        epl2jkz = np.exp(2j * kz_is * dz)
        emn2jkz = np.exp(-2j * kz_is * dz)
           
        # layer response
        L = np.zeros((2,2,2,len_kp), dtype=complex)  # pol, pl/mn1, pl/mn2, kp
        for pol in range(2):
            L[pol,:,:,:] = lay.layersystem_response_matrix(pol, layer_system.thicknesses,
                                                           layer_system.refractive_indices, k_parallel,
                                                           smuthi.fields.angular_frequency(vacuum_wavelength), i_s, i_s)
       
        # transformation coefficients
        B_dag = np.zeros((2, 2, blocksize, len_kp), dtype=complex)  # pol, pl/mn, n, kp
        B = np.zeros((2, 2, blocksize, len_kp), dtype=complex)  # pol, pl/mn, n, kp
        ct = kz_is / k_is
        st = k_parallel / k_is
        _, pilm_pl, taulm_pl = sf.legendre_normalized(ct, st, l_max)
        _, pilm_mn, taulm_mn = sf.legendre_normalized(-ct, st, l_max)
    
        m_list = [None for n in range(blocksize)]
        for tau in range(2):
            for m in range(-m_max, m_max + 1):
                for l in range(max(1, abs(m)), l_max + 1):
                    n = smuthi.fields.multi_to_single_index(tau, l, m, l_max, m_max)
                    m_list[n] = m
                    for pol in range(2):
                        B_dag[pol,0,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                               taulm_list=taulm_pl, dagger=True)
                        B_dag[pol,1,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_mn,
                                                                               taulm_list=taulm_mn, dagger=True)
                        B[pol,0,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                           taulm_list=taulm_pl, dagger=False)
                        B[pol,1,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_mn,
                                                                           taulm_list=taulm_mn, dagger=False)
        
        # pairs of (n1, n2), listed by abs(m1-m2)
        n1n2_combinations = [[] for dm in range(2*m_max+1)]
        for n1 in range(blocksize):
            m1 = m_list[n1]
            for n2 in range(blocksize):
                m2 = m_list[n2]
                n1n2_combinations[abs(m1-m2)].append((n1,n2))
                       
        wr = np.zeros((len_rho, blocksize, blocksize), dtype=complex)
        
        dkp = np.diff(k_parallel)
        if cu.use_gpu:
            re_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.real))
            im_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.imag))
            kernel_source_code = cusrc.radial_lookup_assembly_code %(blocksize, len_rho, len_kp)
            helper_function = cu.SourceModule(kernel_source_code).get_function("helper")
            cuda_blocksize = 128
            cuda_gridsize = (len_rho + cuda_blocksize - 1) // cuda_blocksize
    
            re_dwr_d = cu.gpuarray.to_gpu(np.zeros(len_rho, dtype=np.float32))
            im_dwr_d = cu.gpuarray.to_gpu(np.zeros(len_rho, dtype=np.float32))
        n1n2_combinations = [[] for dm in range(2*m_max+1)]
        for n1 in range(blocksize):
            m1 = m_list[n1]
            for n2 in range(blocksize):
                m2 = m_list[n2]
                n1n2_combinations[abs(m1-m2)].append((n1,n2))
        
        pbar = tqdm(total=blocksize**2, 
                    desc='Layer mediated coupling   ', 
                    file=sys.stdout,
                    bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')
                
        for dm in range(2*m_max+1):
            bessel = scipy.special.jv(dm, (k_parallel[None,:]*radial_distance_array[:,None]))
            besjac = bessel * (k_parallel / (kz_is * k_is))[None,:]
            for n1n2 in n1n2_combinations[dm]:
                n1 = n1n2[0]
                m1 = m_list[n1]
                n2 = n1n2[1]
                m2 = m_list[n2]
                
                belbe = np.zeros(len_kp, dtype=complex)  # n1, n2, kp
                for pol in range(2):
                    belbe += L[pol,0,0,:] * B_dag[pol,0,n1,:] * B[pol,0,n2,:]
                    belbe += L[pol,1,0,:] * B_dag[pol,1,n1,:] * B[pol,0,n2,:] * emn2jkz
                    belbe += L[pol,0,1,:] * B_dag[pol,0,n1,:] * B[pol,1,n2,:] * epl2jkz
                    belbe += L[pol,1,1,:] * B_dag[pol,1,n1,:] * B[pol,1,n2,:]
                
                if cu.use_gpu:
                    re_belbe_d = cu.gpuarray.to_gpu(np.float32(belbe[None, :].real))
                    im_belbe_d = cu.gpuarray.to_gpu(np.float32(belbe[None, :].imag))
    
                    re_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac.real))
                    im_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac.imag))
                    helper_function(re_besjac_d.gpudata, im_besjac_d.gpudata, 
                                    re_belbe_d.gpudata, im_belbe_d.gpudata, 
                                    re_dkp_d.gpudata, im_dkp_d.gpudata, 
                                    re_dwr_d.gpudata, im_dwr_d.gpudata, 
                                    block=(cuda_blocksize, 1, 1), grid=(cuda_gridsize, 1))
                    
                    wr[:,n1,n2] = 4 * (1j)**abs(m2-m1) * (re_dwr_d.get() + 1j*im_dwr_d.get()) 
    
                else:
                    integrand = besjac * belbe[None, :]  # rho, kp 
                    wr[:,n1,n2] = 2 * (1j)**abs(m2-m1) * ((integrand[:,:-1] + integrand[:,1:]) 
                                                                * dkp[None,:]).sum(axis=-1)  # trapezoidal rule
                pbar.update()
        pbar.close()
        w = w + wr
    else:
        sys.stdout.write('Omitting calculations of layer mediated coupling since no difference in layers was detected. \n')
        sys.stdout.flush()
        
    return w , radial_distance_array


def size_format(b):
    if b < 1000:
              return '%i' % b + 'B'
    elif 1000 <= b < 1000000:
        return '%.1f' % float(b/1000) + 'KB'
    elif 1000000 <= b < 1000000000:
        return '%.1f' % float(b/1000000) + 'MB'
    elif 1000000000 <= b < 1000000000000:
        return '%.1f' % float(b/1000000000) + 'GB'
    elif 1000000000000 <= b:
        return '%.1f' % float(b/1000000000000) + 'TB'



def radial_direct_pwe_mediated_coupling_lookup_table(vacuum_wavelength, rho_max, l_max, k_is, k_parallel='default', resolution=None):
    """Prepare Sommerfeld integral lookup table to allow for a fast calculation of the coupling matrix by interpolation.
    This function is called when all particles are on the same z-position.
    
    This function uses PVWF representation of the coupling operator and is only
    for the direct particle-particle lookup table for particles in close vicinity. 
    
    
    Args:
        vacuum_wavelength (float):  Vacuum wavelength in length units
        rho_max (float):            Maximal horizontal particle displacement for which 
                                    direct coupling is evaluated via PVWFs
        l_max (int):                Maximal multipole degree
        k_is (complex):             Wavenumber of the scattering layer
        k_parallel (numpy.ndarray or str):           In-plane wavenumber for Sommerfeld integrals.
                                                     If 'default', smuthi.fields.default_Sommerfeld_k_parallel_array
        resolution (float): Spatial resolution of lookup table in length units. (default: vacuum_wavelength / 100)       
                            Smaller means more accurate but higher memory footprint 
                            
    Returns:
        (tuple) tuple containing:
        
            lookup_table (ndarray):  Coupling lookup, indices are [rho, n1, n2].
            rho_array (ndarray):     Values for the radial distance considered for the lookup (starting from negative 
                                     numbers to allow for simpler cubic interpolation without distinction of cases 
                                     at rho=0)
    """

    sys.stdout.write('Prepare radial particle coupling lookup:\n')
    sys.stdout.flush()

    if resolution is None:
        resolution = vacuum_wavelength / 100
        sys.stdout.write('Setting lookup resolution to %f\n'%resolution)
        sys.stdout.flush()
    
    m_max = l_max # necessary for pwe coupling
    blocksize = smuthi.fields.blocksize(l_max, m_max)
    
    radial_distance_array = np.arange(- 3 * resolution, rho_max + 3 * resolution, resolution)   
    len_rho = len(radial_distance_array)
    
    # pwe-coupling, similar to layer mediated coupling -------------------------------------------------------------------------------------
    sys.stdout.write('PWE-coupling correction for direct coupling: ...')
    sys.stdout.flush()
    
    if type(k_parallel) == str and k_parallel == 'default':
        k_parallel = smuthi.fields.default_Sommerfeld_k_parallel_array

    kz_is = smuthi.fields.k_z(k_parallel=k_parallel, k=k_is)
    len_kp = len(k_parallel)

       
    # transformation coefficients
    B_dag = np.zeros((2, blocksize, len_kp), dtype=complex)  # pol, n, kp (pl only)
    B = np.zeros((2, blocksize, len_kp), dtype=complex)  # pol, n, kp (pl only)
    ct = kz_is / k_is
    st = k_parallel / k_is
    _, pilm_pl, taulm_pl = sf.legendre_normalized(ct, st, l_max)
    _, pilm_mn, taulm_mn = sf.legendre_normalized(-ct, st, l_max)

    m_list = [None for n in range(blocksize)]
    for tau in range(2):
        for m in range(-m_max, m_max + 1):
            for l in range(max(1, abs(m)), l_max + 1):
                n = smuthi.fields.multi_to_single_index(tau, l, m, l_max, m_max)
                m_list[n] = m
                for pol in range(2):
                    B_dag[pol,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                           taulm_list=taulm_pl, dagger=True)
                    B[pol,n,:] = trf.transformation_coefficients_vwf(tau, l, m, pol, pilm_list=pilm_pl,
                                                                       taulm_list=taulm_pl, dagger=False)

    
    # pairs of (n1, n2), listed by abs(m1-m2)
    n1n2_combinations = [[] for dm in range(2*m_max+1)]
    for n1 in range(blocksize):
        m1 = m_list[n1]
        for n2 in range(blocksize):
            m2 = m_list[n2]
            n1n2_combinations[abs(m1-m2)].append((n1,n2))
    
    # in the rotated coordinate system, the radial distance rho is represented 
    # as a vertical distance dz
    dz_array = radial_distance_array 
    epljkdz = np.exp(1j * kz_is[None, :] * dz_array[:, None])               
    w = np.zeros((len_rho, blocksize, blocksize), dtype=complex)
    
    dkp = np.diff(k_parallel)
    if cu.use_gpu:
        re_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.real))
        im_dkp_d = cu.gpuarray.to_gpu(np.float32(dkp.imag))
        kernel_source_code = cusrc.radial_lookup_assembly_code_PVWF %(blocksize, len_rho, len_kp)
        helper_function = cu.SourceModule(kernel_source_code).get_function("helper")
        cuda_blocksize = 128
        cuda_gridsize = (len_rho + cuda_blocksize - 1) // cuda_blocksize

        re_dw_d = cu.gpuarray.to_gpu(np.zeros(len_rho, dtype=np.float32))
        im_dw_d = cu.gpuarray.to_gpu(np.zeros(len_rho, dtype=np.float32))
    n1n2_combinations = [[] for dm in range(2*m_max+1)]
    for n1 in range(blocksize):
        m1 = m_list[n1]
        for n2 in range(blocksize):
            m2 = m_list[n2]
            n1n2_combinations[abs(m1-m2)].append((n1,n2))
    
    pbar = tqdm(total=blocksize**2, 
                desc='PWE direct coupling   ', 
                file=sys.stdout,
                bar_format='{l_bar}{bar}| elapsed: {elapsed} remaining: {remaining}')
            
    for dm in range(2*m_max+1):
        # in the rotated coordinate system, rho is always zero
        bessel = scipy.special.jv(dm, (k_parallel[None,:]*np.zeros(len_rho)[:,None]))
        besjac = bessel * (k_parallel / (kz_is * k_is))[None,:]
        for n1n2 in n1n2_combinations[dm]:
            n1 = n1n2[0]
            m1 = m_list[n1]
            n2 = n1n2[1]
            m2 = m_list[n2]
            
            bebe = np.zeros((len_rho, len_kp), dtype=complex) # n1, n2, rho, kp
            for pol in range(2):
                # always use B+ (all kz > 0)
                bebe += ((B_dag[pol, n1, :] * B[pol, n2, :])[None, :] * epljkdz) 
                
            if cu.use_gpu:
                re_bebe_d = cu.gpuarray.to_gpu(np.float32(bebe.real)) 
                im_bebe_d = cu.gpuarray.to_gpu(np.float32(bebe.imag)) 

                re_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac.real))
                im_besjac_d = cu.gpuarray.to_gpu(np.float32(besjac.imag))
                helper_function(re_besjac_d.gpudata, im_besjac_d.gpudata, 
                                re_bebe_d.gpudata, im_bebe_d.gpudata, 
                                re_dkp_d.gpudata, im_dkp_d.gpudata, 
                                re_dw_d.gpudata, im_dw_d.gpudata, 
                                block=(cuda_blocksize, 1, 1), grid=(cuda_gridsize, 1))
                
                w[:,n1,n2] = 4 * (1j)**abs(m2-m1) * (re_dw_d.get() + 1j*im_dw_d.get()) 

            else:
                integrand = besjac * bebe  # rho, kp 
                w[:,n1,n2] = 2 * (1j)**abs(m2-m1) * ((integrand[:,:-1] + integrand[:,1:]) 
                                                            * dkp[None,:]).sum(axis=-1)  # trapezoidal rule
            pbar.update()
    pbar.close()
    
    alpha = 0
    beta = -np.pi / 2
    rot_mat_1 = np.transpose(trf.block_rotation_matrix_D_svwf(l_max, m_max, 0, beta, alpha))
    rot_mat_2 = np.transpose(trf.block_rotation_matrix_D_svwf(l_max, m_max, -alpha, -beta, 0))
    w = np.dot(np.einsum('ij,ljk->lik', rot_mat_1, w), rot_mat_2)
    
    return w, radial_distance_array
